/*
 * linux_mpeg2: drivers for the Visiontech/ BMK Kfir MPEG2 encoder card
 * Copyright (C) 1999,2000 convergence integrated media GmbH
 * 

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 * 

 * The author can be reached at ralph@convergence.de, 

 * the project's page is at http://linuxtv.org/mpeg2/
 */

#define __KERNEL_SYSCALLS__
#include <linux/poll.h>
#include <linux/fs.h>
#include <linux/unistd.h>
#include <linux/fcntl.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/pci.h>
#include <linux/wrapper.h>
#include <linux/init.h>
#include <linux/ctype.h>
#include <asm/io.h>
#if defined(MODVERSIONS)
#include <linux/modversions.h>
#endif

#ifdef __KFIR_PACK__
#include "videodev.h"
#include "i2c.h"
#include "i2c-algo-bit.h"
#include "video_decoder.h"
#else
#include <linux/videodev.h>
#include <linux/i2c.h>
#include <linux/i2c-algo-bit.h>
#include <linux/video_decoder.h>
#endif

#include "micro/kfir_micro.h"
#include "kfir.h"
#include "plx9054.h"

MODULE_AUTHOR("Ralph Metzler <ralph@convergence.de>");
MODULE_DESCRIPTION("Visiontech Kfir MPEG2 encoder driver");
MODULE_PARM(debug,"i");
MODULE_PARM(loaded,"i");

static int debug = 0;
static int loaded = 0;


MODULE_PARM(vidinmode,"i");
MODULE_PARM(vidinput,"i");
MODULE_PARM(vidoutmode,"i");
MODULE_PARM(vidsize,"i");
MODULE_PARM(vidrate,"i");
MODULE_PARM(mpeg1mode,"i");
MODULE_PARM(streamtype,"i");
MODULE_PARM(sndfreq,"i");

static int vidinmode = VIDEO_MODE_PAL;
static int vidinput = 0;
static int vidoutmode = VT_KFIR_PAL;
static int vidsize = (KFIR_V_576*16+KFIR_H_FD1);
static int vidrate = 4500000;
static int sndfreq = VT_KFIR_AUDIO_FREQ_48_000;
static int mpeg1mode = 0;
static int streamtype = VT_KFIR_PROGRAM;


#define dprintk	if (debug) printk

static struct kfir_dev kfirs[MAX_NUM_KFIR];
static int num_kfir = 0;

// Define to allow burst mode with memory mapping
//#define ALTERA_BURST  1
#define PLX_BURST   1


/* Not a full fletched 64 bit routine but should be accurate enough for what we need */
inline u64 div64(u32 a, u32 b, u32 m)
{
	u32 r, s;
	
	if (!b)
		return 0;
	r=(a/b);
	s=((a%b)*m)/b;
	
	return (((u64)r)*m)|s;
}

#define RMUL(a,b,c) div64((a),(c),(b))


static inline void ddelay(int i) 
{
        current->state=TASK_INTERRUPTIBLE;
        schedule_timeout((HZ*i)/100);
}

static inline void Sleep(int i) 
{
	if (i>10)
		ddelay((9+i)/10);
	else
		mdelay(i);
}

static int errno;
static int do_firmread(const char *fn, char **fp)
{
        int fd;
	long l;
	char *dp;
        
	fd = open(fn, 0, 0);
	if (fd == -1)
	{
		printk(KERN_INFO "Unable to load '%s'.\n", fn);
		return 0;
	}
	l = lseek(fd, 0L, 2);
	if (l <= 0 || l > 300*1024)
	{
	        printk(KERN_INFO "Firmware too large '%s'\n", fn);
		sys_close(fd);
		return 0;
	}
	lseek(fd, 0L, 0);
	*fp= dp = vmalloc(l);
	if (dp == NULL)
	{
		printk(KERN_INFO "Out of memory loading '%s'.\n", fn);
		sys_close(fd);
		return 0;
	}
	if (read(fd, dp, l) != l)
	{
		printk(KERN_INFO "Failed to read '%s'.\n", fn);
		vfree(dp);
		sys_close(fd);
		return 0;
	}
	close(fd);
	*fp = dp;
	return (int) l;
}

static
int firmread(const char *fn, char **fp)
{
	int r;
	mm_segment_t fs = get_fs();
        
	set_fs(get_ds());
	r = do_firmread(fn, fp);
	set_fs(fs);
	return r;
}

/* ----------------------------------------------------------------------- */
/* ADSP functions                                                          */

static u16 ADSP_MonoRates[10]   = {32,48,56,64,80,96,112,128,160,192};
static u16 ADSP_StereoRates[10] = {64,96,112,128,160,192,224,256,320,384};
static u16 ADSP_AllRates[14]    = {32,48,56,64,80,96,112,128,160,192,224,256,320,384};
static u16 ADSP_FreqSample[8]   = {0x500,0x700,0x400,0x600,0x300,0x200,0x000,0x100};

static const WORD*
ADSP_GetRates(struct kfir_dev *kfir, WORD AudioMode, WORD AudioFormat)
{
        if (!AudioFormat) // Layer I
	        return ADSP_AllRates;

	// Layer II
	if (!AudioMode)
		return ADSP_MonoRates;
	else
		return ADSP_StereoRates;
}

static WORD 
ADSP_SendCommand(struct kfir_dev *kfir, const WORD Command)
{
	u32 start;
	WORD wd=0x8000;
	
	ADSPWriteAddr(ADR4);
	ADSPWriteData(0xffff);

	ADSPWriteAddr(ADR0);
	mdelay(1);
	ADSPWriteData(Command);
        udelay(500);

	ADSPWriteIRQ(0);
	udelay(10);
	ADSPWriteIRQ(2);
	//ADSP_SendSystemCommand(kfir, SYSCMD_MASTERDSPINT);

	start=jiffies;
	while(jiffies-start <= HZ) {  
		Sleep(3);
		ADSPWriteAddr(ADR4);
		wd=ADSPReadData;
		if (wd != 0xFFFF) {
			ADSPWriteAddr(ADR4);
			wd=ADSPReadData;
			break;
		}
	}
	
	if ((wd & 0x8000)==0)
		return wd;
	
	return 0x8000;
}

static int ADSP_SetupPLL(struct kfir_dev *kfir,  const WORD AudioFreq)
{
	DWORD Data;
	
	switch (AudioFreq) {
	case 0: // 5.5125KHz
	case 1: // 8KHz
	case 2: // 11.025KHz
	case 3: // 16KHz
	case 4: // 22.5KHz
		return -1;
	case 5: // 32KHz
		Data = 0;
		break;
	case 6: // 44.1KHz
		Data = 0x5;
		break;
	default:
	case 7: // 48KHz
		Data = 0x6;
		break;
	}
	waltl(ALT_ADR_ADSP_FREQ, Data);
	return 0;
}


static int ADSP_CodeLoad(struct kfir_dev *kfir)
{
	u16 *CodePtr;
	u16 *Boundary;
	u16 Data, BlockSize, adr;

	u32 microlen;
	u16 *micro;
	
	if (
#ifdef __KFIR_PACK__
	    !(microlen=firmread("micro/MPEG1_6.HLD", (char**) &micro)) &&
#endif
	    !(microlen=firmread(KFIR_MICRO_PATH "MPEG1_6.HLD", (char**) &micro)))
		return -1;

	CodePtr = micro+5;
	Boundary= CodePtr+micro[1];

	ADSPWriteAddr(0x5000);
	ADSPWriteData(0xAA55);

	ADSPWriteAddr(0x5000);
	Data = ADSPReadData;

	while(CodePtr < Boundary) {
		BlockSize = *(CodePtr++);
		adr = *(CodePtr++);
		//printk("ADSP: write %04x bytes to %04x\n", BlockSize, adr);
		ADSPWriteAddr(adr);
		while(BlockSize-- && CodePtr < Boundary) 
			ADSPWriteData(*(CodePtr++));
	}
	vfree(micro);
	udelay(5);
	
	ADSPWriteAddr(0x1);
	Data = ADSPReadData;
	
	ADSPWriteAddr(0x1);
	ADSPWriteData(0xAA55);
	
	ADSPWriteAddr(0x1);
	Data = ADSPReadData;
  
	ADSPWriteAddr(0x5000);
	udelay(3);
	ADSPWriteData(0x5555);
	
	udelay(3);

	ADSPWriteAddr(0x5000);
	udelay(3);
	if (ADSPReadData != 0x5555) {
		printk("kfir: ADSP loading failed!\n");
		return -1;
	}
	printk("kfir: ADSP loading successful\n");
	return 0;
}


static int ADSP_KfInit(struct kfir_dev *kfir)
{
        int eRet;
	int nRetries = 0;
	const WORD* RateArray=ADSP_GetRates(kfir, 
					    kfir->Params.AudioMode, 
					    kfir->Params.AudioFormat);
	DWORD Data;
	
	kfir->m_CmdReg = 0xf;
	if (!kfir->FlValid) {
	        Data=raltl(ALT_ADR_ADSP_IRQ);
		Data &= ~(ALT_D_ADSP_IRQ_E | ALT_D_ADSP_IRQ_2);
		waltl(ALT_ADR_ADSP_IRQ, Data);
		Sleep(3);
		Data |= ALT_D_ADSP_IRQ_E | ALT_D_ADSP_IRQ_2;
		waltl(ALT_ADR_ADSP_IRQ, Data);
		
		do {
		        if (!ADSP_CodeLoad(kfir)) {
			        kfir->FlValid = 1;
			        break;         
			}
		} while (nRetries++ <= 8);
		
		if (nRetries == 8) {
		        kfir->FlValid = 0;
			return -1;
		}
	}
	
	Sleep(10); 
	kfir->FlValid = 0;
	eRet = -1;
	nRetries = 0;
	do {
	        WORD wd = 0;
		printk("kfir: ADSP initialization...");
		
		ADSP_SetupPLL(kfir, kfir->Params.AudioFreq);
		Sleep(100);
		
		// Reset ADSP.
		
		// Pulse IRQ E and IRQ 2
		Data=raltl(ALT_ADR_ADSP_IRQ);
		Data &= ~(ALT_D_ADSP_IRQ_E | ALT_D_ADSP_IRQ_2);
		waltl(ALT_ADR_ADSP_IRQ, Data);
		
		Data |= ALT_D_ADSP_IRQ_E | ALT_D_ADSP_IRQ_2;
		Data=waltl(ALT_ADR_ADSP_IRQ, Data);
		
		// Some magic
		ADSPWriteAddr(0);
		ADSPWriteData(0x3C0F);
		ADSPWriteData(0x00FC);
		Sleep(10);   // Critical !!!!
		
		// Initialize ADSP.
		wd|=ADSP_SendCommand(kfir, 
				      DSPAPI_SETSTYLE | 
				      ((kfir->Params.AudioFormat+0x41) << 8));
		wd|=ADSP_SendCommand(kfir, DSPAPI_SETSAMPLINGRATE | 
				      ADSP_FreqSample[kfir->Params.AudioFreq]);
		wd|=ADSP_SendCommand(kfir, DSPAPI_SETCRCOPTIONS |
				      ((2-kfir->Params.AudioCRC) << 8));
		wd|=ADSP_SendCommand(kfir, DSPAPI_SETCHANNELS | 
				      (kfir->Params.AudioMode << 8));
		wd|=ADSP_SendCommand(kfir, 
				      DSPAPI_SETRESOLUTION | 
				      ((RateArray[kfir->Params.AudioBitRate]/2) << 8));
		wd|=ADSP_SendCommand(kfir, DSPAPI_QUERYSUPPORTRECORD);
		
		if (!(wd & 0x8000)) {
		        kfir->FlValid = 1;
			eRet = 0;
			printk("done\n");
			break;
		}
		printk("failed\n");
		Sleep(5);
	} while(nRetries++ < 5);
	
	return eRet;
}

static bool ADSP_Start(struct kfir_dev *kfir)
{
        WORD wd=0;
	
	kfir->m_CmdReg = 0xf;
	wd = ADSP_SendCommand(kfir, DSPAPI_STARTRECORD);

	if (wd & 0x8000)
	        return -1;
	return 0;
}

/* ----------------------------------------------------------------------- */
/* Kfir functions                                                          */

static int Kfir_LoadMuxParams(struct kfir_dev *kfir)
{
	int i;

	for (i = 0; i < MAX_MUX_PARAMS; i++) {
		KfirRegWrite(KFIR_202, (kfir->MuxParams[i]>>16)&0xFFFF);
		KfirRegWrite(KFIR_204, kfir->MuxParams[i] & 0xFFFF);
	}
	return 0;
}


u8 PAT[] = {
	0x00, 0x00, /* PAT TID */
	0x00, 0x00, /* PAT PID */
	0x00, 0x14, /* size */
	
	0x00, 0xb0, 0x11, 
	0x00, 0x00, 0xcb, 0x00, 0x00, 0x00, 0x00,
	0xe0, 0x10, 0x00, 0x01, 0xe4, 0x00, 
	0x2a, 0xd6, 0x1a, 0xfb,

	0x00, 0x02, /* PMT TID */
	0x04, 0x00, /* PAT PID */
	0x00, 0x1f, /* size */
	
	0x02, 0xb0, 0x1c,
	0x00, 0x01, 0xcb, 0x00, 0x00, 0xe0, 0xa0, 0xf0, 
	0x05, 0x48, 0x03, 0x01, 0x00, 0x00, 0x02, 0xe0, 
	0xa0, 0xf0, 0x00, 0x03, 0xe0, 0x50, 0xf0, 0x00,
	0xae, 0xea, 0x4e, 0x48,
	0x00,

	0x00, 0x00, 
	0x00, 0x00, 
	0x00, 0x00, 
};


static void
CreateTables(struct kfir_dev *kfir)
{
	int size=70, i;
	
	for (i = 0  ; i < (size / 2) ; i++)
		kfir->MuxParams[MUX_AUDIO_TABLES+i]=(PAT[2*i]<<8) | (PAT[2*i+1]);

	kfir->MuxParams[MUX_AUDIO_ADR_PAT_TABLE_SIZE] = 26;
	kfir->MuxParams[MUX_AUDIO_ADR_PMT_TABLE_SIZE] = 44;
}

static int Kfir_LoadAudioTable(struct kfir_dev *kfir, const DWORD AudioMemBase)
{
        int i;
	DWORD Data;
	QWORD Scale;
	BYTE FlPesScramblingCntrl = 0;
	BYTE FlPesPriority = 0;
	BYTE FlDataAlignment = 0;
	BYTE Copyright = 0; //gui
	BYTE FlOriginal = 0; //gui
	BYTE FlDSMTrickMode = 0;
	BYTE FlAdditionalCopy = 0;
	BYTE FlPesCrcFlag = 0;
	BYTE FlPesExtension = 0;
	BYTE FlDiscontinuity = 0;
	BYTE FlRnd = 0;
	BYTE FlEPriority = 0;
	BYTE FlOPCR = 0;
	BYTE FlSplicing = 0;
	BYTE FlPrv = 0;
	BYTE Extension = 0;
	DWORD AudioBound = 1;
	BYTE FlFIxed = 1;
	BYTE FlCsps = 0;
	BYTE FlSystemAudioLock = 1;
	BYTE FlSystemVideoLock = 1;
	DWORD VideoBound = 1;
	BYTE FlPacketrateRestriction = 1;
	DWORD AudioBoundScale = 0;
	DWORD AudioBufferSize = 0x50;
	DWORD VideoBoundScale = 1;
	DWORD VideoBufferSize = 0x50;
	DWORD StreamType;

	const WORD* RateArray=ADSP_GetRates(kfir, kfir->Params.AudioMode, 
					    kfir->Params.AudioFormat);
	
	for (i = 0 ; i < MAX_MUX_PARAMS ; i++) {
		kfir->MuxParams[ i ] = 0;
	}

	switch(kfir->Params.StreamType) {
	case VT_KFIR_AUDIO_ES:
		StreamType = KFIR_D_AUDIO_ES;
		break;
	case VT_KFIR_SYSTEM:
		StreamType = KFIR_D_SYSTEM;
		break;
	case VT_KFIR_VIDEO_ES:
	case VT_KFIR_A_V_ES:
		StreamType = KFIR_D_VEDEO_ES;
		break;
	case VT_KFIR_PROGRAM:
		StreamType = KFIR_D_PROGRAM;
		break;  
	case VT_KFIR_TRANSPORT:
		StreamType = KFIR_D_TRANSPORT;
		break;
	case VT_KFIR_PACKET_VES:
		StreamType = KFIR_D_PACKET_VIDEO_ES;
		break;
	case VT_KFIR_PACKET_AES:
		StreamType = KFIR_D_PACKET_AUDIO_ES;
		break;
	default:
		StreamType = KFIR_D_PROGRAM;
		break;
	}

	kfir->MuxParams[MUX_AUDIO_ADR_STREAM_TYPE] = StreamType;
	kfir->MuxParams[MUX_AUDIO_ADR_PROGRAM_NUMBER] = kfir->Params.ProgramNumber;  
	kfir->MuxParams[MUX_AUDIO_ADR_TRANSPORT_STREAM_ID] = kfir->Params.TransportStreamID;
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_STREAM_TYPE] = DEFAULT_AUDIO_STREAM_TYPE;
	kfir->MuxParams[MUX_AUDIO_ADR_PMT_PID] = kfir->Params.ProgramMapID&0x1fff;
	kfir->MuxParams[MUX_AUDIO_ADR_VIDEO_PID] = kfir->Params.VideoProgramID&0x1fff;
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_PID] = kfir->Params.AudioProgramID&0x1fff;
	kfir->MuxParams[MUX_AUDIO_ADR_VIDEO_STREAM_ID]=kfir->Params.VideoStreamID&0xff;
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_STREAM_ID]=kfir->Params.AudioStreamID&0xff;
	Data = kfir->Params.BitRate;  // in bits !!!
	kfir->MuxParams[MUX_AUDIO_ADR_VIDEO_BITRATE_LOW] = Data & 0xffff;
	kfir->MuxParams[MUX_AUDIO_ADR_VIDEO_BITRATE_HIGH] = (Data>>16) & 0xffff;
	
	Data = RateArray[kfir->Params.AudioBitRate] * 1000;
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_BITRATE_LOW] = Data & 0xffff;
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_BITRATE_HIGH] = (Data>>16) & 0xffff;
	kfir->MuxParams[MUX_AUDIO_ADR_TRANSPORT_BITRATE_LOW] = kfir->Params.TransportRate & 0xffff;
	kfir->MuxParams[MUX_AUDIO_ADR_TRANSPORT_BITRATE_HIGH] = (kfir->Params.TransportRate >> 16) & 0xffff;
	kfir->MuxParams[MUX_AUDIO_ADR_VPTS_OFFSET_LOW] = kfir->Params.InitialVideoPTS & 0xffff;
	kfir->MuxParams[MUX_AUDIO_ADR_VPTS_OFFSET_HIGH] = (kfir->Params.InitialVideoPTS >> 16) & 0xffff;
	kfir->MuxParams[MUX_AUDIO_ADR_APTS_OFFSET_LOW] = kfir->Params.InitialAudioPTS & 0xffff;
	kfir->MuxParams[MUX_AUDIO_ADR_APTS_OFFSET_HIGH] = (kfir->Params.InitialAudioPTS >> 16) & 0xffff;
	kfir->MuxParams[MUX_AUDIO_ADR_MAX_VIDEO_PES_SIZE] = DEFAULT_VIDEO_PES_SIZE; 
	kfir->MuxParams[MUX_AUDIO_ADR_MAX_AUDIO_PES_SIZE] = DEFAULT_AUDIO_PES_SIZE; 
	kfir->MuxParams[MUX_AUDIO_ADR_PAT_RATE] = DEFAULT_PAT_RATE; 
	kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC] =
		kfir->Params.Video_Format == VT_KFIR_NTSC ? 2997 : 2500;

	// Additional parameters for 3/2 pulldown
	kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC_SEC] = 2400; 
	
	Scale = div64(kfir->Params.TransportRate * 100,
		      kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC], 1<<13);

	kfir->MuxParams[MUX_AUDIO_ADR_PCR_BITLIMIT_LOW] = (DWORD)(Scale & 0xffff);
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_BITLIMIT_MID] = (DWORD)((Scale >> 16) & 0xffff);
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_BITLIMIT_HIGH] = (DWORD)((Scale >> 32) & 0xffff);
	
	// Additional parameters for 3/2 pulldown
	Scale = div64(kfir->Params.TransportRate * 100,
		      kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC_SEC], 1<<13);

	kfir->MuxParams[MUX_AUDIO_ADR_PCR_BITLIMIT_LOW_SEC] = (DWORD)(Scale & 0xffff);
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_BITLIMIT_MID_SEC] = (DWORD)((Scale >> 16) & 0xffff);
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_BITLIMIT_HIGH_SEC] = (DWORD)((Scale >> 32) & 0xffff);
	
	Scale=div64(27000000UL * 100,    // MUX_STEP_SHIFT
		    kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC], 1<<16);

	kfir->MuxParams[MUX_AUDIO_ADR_PCR_STEP_LOW] = (DWORD)(Scale & 0xffff);
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_STEP_MID] = (DWORD)((Scale >> 16) & 0xffff);
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_STEP_HIGH] = (DWORD)((Scale >> 32) & 0xffff); 
	
	// Additional parameters for 3/2 pulldown
	Scale=div64(27000000UL * 100,    // MUX_STEP_SHIFT
		    kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC_SEC], 1<<16);

	kfir->MuxParams[MUX_AUDIO_ADR_PCR_STEP_LOW_SEC] = (DWORD)(Scale & 0xffff);
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_STEP_MID_SEC] = (DWORD)((Scale >> 16) & 0xffff);
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_STEP_HIGH_SEC] = (DWORD)((Scale >> 32) & 0xffff); 
	
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_STEP_100SEC_LOW] = ((QWORD)27000000UL * 100L) & 0xffff;
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_STEP_100SEC_HIGH] = ((QWORD)27000000UL * 100 >> 16) & 0xffff;
	
	Scale=div64(90000UL * 100,    // MUX_STEP_SHIFT
		    kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC], 1<<16);

	kfir->MuxParams[MUX_AUDIO_ADR_PCRBASE_STEP_LOW] = (DWORD)(Scale & 0xffff);
	kfir->MuxParams[MUX_AUDIO_ADR_PCRBASE_STEP_HIGH] = (DWORD)((Scale >> 16) & 0xffff);
	
	// Additional parameters for 3/2 pulldown
	Scale=div64(90000UL * 100,    // MUX_STEP_SHIFT
		    kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC_SEC], 1<<16);

	kfir->MuxParams[MUX_AUDIO_ADR_PCRBASE_STEP_LOW_SEC] = (DWORD)(Scale & 0xffff);
	kfir->MuxParams[MUX_AUDIO_ADR_PCRBASE_STEP_HIGH_SEC] = (DWORD)((Scale >> 16) & 0xffff);
	
	kfir->MuxParams[MUX_AUDIO_ADR_PCRBASE_STEP_100SEC_LOW] = (90000UL * 100) & 0xffff;
	kfir->MuxParams[MUX_AUDIO_ADR_PCRBASE_STEP_100SEC_HIGH] = ((90000UL  * 100) >> 16) & 0xffff;
	
	kfir->MuxParams[MUX_AUDIO_ADR_PPH_C1] = 2 << 6 | FlPesScramblingCntrl << 4 | FlPesPriority << 3 | 
		FlDataAlignment << 2 | Copyright << 1 | FlOriginal;
	kfir->MuxParams[MUX_AUDIO_ADR_PPH_C2] = FlDSMTrickMode << 3 | FlAdditionalCopy << 2 | FlPesCrcFlag << 1 | FlPesExtension;
	kfir->MuxParams[MUX_AUDIO_ADR_TPH_C1] = FlDiscontinuity << 2 | FlRnd << 1 | FlEPriority;
	kfir->MuxParams[MUX_AUDIO_ADR_TPH_C2] = FlOPCR << 3 | FlSplicing << 2 | FlPrv << 1 | Extension;
	
	kfir->MuxParams[MUX_AUDIO_ADR_ES_RATE_FLAG] = 0;
	kfir->MuxParams[MUX_AUDIO_ADR_ESCR_FLAG] = 0;
	kfir->MuxParams[MUX_AUDIO_ADR_ESCR_BASE_LOW] = 0;
	kfir->MuxParams[MUX_AUDIO_ADR_ESCR_BASE_MID] = 0;
	kfir->MuxParams[MUX_AUDIO_ADR_ESCR_BASE_HIGH] = 0;
	kfir->MuxParams[MUX_AUDIO_ADR_ESCR_EXTENTION] = 0;
	
	kfir->MuxParams[MUX_AUDIO_ADR_SCRAMBLING] = 0;
	kfir->MuxParams[MUX_AUDIO_ADR_ERROR_INDICATOR] = 0;
	kfir->MuxParams[MUX_AUDIO_ADR_PRIORITY] = 0;
	
	kfir->MuxParams[MUX_AUDIO_ADR_DTS_PER_FRAME] = 90000UL * 100 / kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC];
	// Additional parameters for 3/2 pulldown
	kfir->MuxParams[MUX_AUDIO_ADR_DTS_PER_FRAME_SEC] = 90000UL * 100 / kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC_SEC];
	
	Scale=div64(90000UL * 100,    // MUX_STEP_SHIFT
		    kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC],
		    kfir->Params.KF_GOP_M);

	kfir->MuxParams[MUX_AUDIO_ADR_PTS_DTS_OFFSET] = (DWORD)(Scale & 0xffff);
	// Additional parameters for 3/2 pulldown
	Scale=div64(90000UL * 100,    // MUX_STEP_SHIFT
		    kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC_SEC],
		    kfir->Params.KF_GOP_M);

	kfir->MuxParams[MUX_AUDIO_ADR_PTS_DTS_OFFSET_SEC] = (DWORD)(Scale & 0xffff);
	
	Scale=div64(27000000UL*32, kfir->Params.TransportRate, 1024);
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_CORRECTION_CONST_LOW]=
		(DWORD)(Scale & 0x7ff); 
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_CORRECTION_CONST_HIGH] = 
		(DWORD)((Scale >> 11) & 0x7ff);
	
	Scale=div64(90000UL * 32, kfir->Params.TransportRate, 1024);
	kfir->MuxParams[MUX_AUDIO_ADR_PCRBASE_CORRECTION_CONST_LOW] = (DWORD)(Scale & 0x7ff);
	kfir->MuxParams[MUX_AUDIO_ADR_PCRBASE_CORRECTION_CONST_HIGH] = (DWORD)((Scale >> 11) & 0x7ff);
	
	Data = (kfir->Params.BitRate+(RateArray[kfir->Params.AudioBitRate]*1000))/400;
	kfir->MuxParams[MUX_AUDIO_ADR_PROGRAM_MUX_RATE_LOW]=Data & 0xffff; 
	kfir->MuxParams[MUX_AUDIO_ADR_PROGRAM_MUX_RATE_HIGH]=(Data >> 16) & 0x3f;
	kfir->MuxParams[MUX_AUDIO_ADR_RATE_BOUND_LOW]=Data & 0xffff;
	kfir->MuxParams[MUX_AUDIO_ADR_RATE_BOUND_HIGH]=(Data >> 16) & 0x3f;
	
	Data=(kfir->Params.BitRate+(RateArray[kfir->Params.AudioBitRate]*1000))/400;
	kfir->MuxParams[MUX_AUDIO_ADR_ES_RATE_LOW] = Data & 0xffff;
	kfir->MuxParams[MUX_AUDIO_ADR_ES_RATE_HIGH] = (Data >> 16) & 0xffff;
	
	Data = (1<<24) | (AudioBound<<18) | (FlFIxed<<17) | (FlCsps<<16) |
		(FlSystemAudioLock<<15) | (FlSystemVideoLock<<14) |
		(1<<13) | (VideoBound<<8) | (FlPacketrateRestriction<<7) | 0x7F;
	kfir->MuxParams[MUX_AUDIO_ADR_SHF_LOW] = Data & 0xffff;
	kfir->MuxParams[MUX_AUDIO_ADR_SHF_HIGH] = (Data >> 16) & 0x1ff;
	
	Data = (kfir->Params.AudioStreamID << 16) |
		(3 << 14) | (AudioBoundScale << 13) | AudioBufferSize;
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_PAR_CONST_LOW] = Data & 0xffff;
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_PAR_CONST_HIGH] = 
		(Data >> 16) & 0xffff;
	Data = (kfir->Params.VideoStreamID << 16) | (3 << 14) |
		(VideoBoundScale << 13) | VideoBufferSize;
	kfir->MuxParams[MUX_AUDIO_ADR_VIDEO_PAR_CONST_LOW] = Data & 0xffff;
	kfir->MuxParams[MUX_AUDIO_ADR_VIDEO_PAR_CONST_HIGH]=(Data>>16)&0xffff;
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_MISS_LATENCY] = 15; // gui
	
	// NEW Parameters
	Data = 90000 * 100;
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_MISS_LATENCY+1]=(Data&0xffff);
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_MISS_LATENCY+2]=((Data>>16)&0xffff);
	
	
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_PUSH] = 1;

	switch(kfir->Params.AudioFreq) {
	case VT_KFIR_AUDIO_FREQ_5_5125:
	case VT_KFIR_AUDIO_FREQ_8_0000:
	case VT_KFIR_AUDIO_FREQ_11_025:
	case VT_KFIR_AUDIO_FREQ_16_000:
	case VT_KFIR_AUDIO_FREQ_22_050:
	case VT_KFIR_AUDIO_FREQ_44_100:
	default:
		kfir->MuxParams[MUX_AUDIO_ADR_FRAME_SIZE] = 
			((1440 * RateArray[kfir->Params.AudioBitRate]) / 441)&~3;
		kfir->MuxParams[MUX_AUDIO_ADR_FRAME_TIME] = 2351;
		break;
	case VT_KFIR_AUDIO_FREQ_32_000:
		kfir->MuxParams[MUX_AUDIO_ADR_FRAME_SIZE] = 
			(144 * RateArray[kfir->Params.AudioBitRate]) / 32;
		kfir->MuxParams[MUX_AUDIO_ADR_FRAME_TIME] = 36 * 90;
		break;
	case VT_KFIR_AUDIO_FREQ_48_000:
		kfir->MuxParams[MUX_AUDIO_ADR_FRAME_SIZE] = 
			(144 * RateArray[kfir->Params.AudioBitRate]) / 48;
		kfir->MuxParams[MUX_AUDIO_ADR_FRAME_TIME] = 24 * 90;
		break;
	}

	CreateTables(kfir);

	return Kfir_LoadMuxParams(kfir);
}


static void ReadKfirStateGeneric(struct kfir_dev *kfir)
{
	  kfir->KfState.OldIntCnt = KfirRegRead(KFIR_ADR_INTERRUPT_CNT);
	  kfir->KfState.IntDMA_Mux = KfirRegRead(KFIR_ADR_INT_DMA_MUX);
	  kfir->KfState.IntVideo = KfirRegRead(KFIR_ADR_INT_VIDEO);
	  kfir->KfState.IntAudio = KfirRegRead(KFIR_ADR_INT_AUDIO);
	  kfir->KfState.IntDMA_VIB = KfirRegRead(KFIR_ADR_INT_DMA_VIB);
	  kfir->KfState.MuxCnt_0 = KfirRegRead(KFIR_ADR_MUX_CNT_0);
	  kfir->KfState.MuxCnt_1 = KfirRegRead(KFIR_ADR_MUX_CNT_1);
	  kfir->KfState.MuxCnt_2 = KfirRegRead(KFIR_ADR_MUX_CNT_2);
	  kfir->TotalLinesAfterMUX = kfir->KfState.MuxCnt_0 + 
		  (kfir->KfState.MuxCnt_1 << 16) + 
		  ((QWORD)(kfir->KfState.MuxCnt_2) << 32);
	  kfir->KfState.GOP_0 = KfirRegRead(KFIR_ADR_GOP_0);
	  kfir->KfState.GOP_1 = KfirRegRead(KFIR_ADR_GOP_1);
	  kfir->KfState.Frame = KfirRegRead(KFIR_ADR_FRAME);
  
	  kfir->KfState.KfirEReady = KfirRegRead(KFIR_ADR_EREADY);
}

static void Kfir_Start(struct kfir_dev *kfir)
{
	DWORD Data;
	printk("eready=%08x\n",KfirRegRead(KFIR_ADR_EREADY));
	Data = KfirRegRead(KFIR_ADR_RECORD);
	KfirRegWrite(KFIR_ADR_RECORD, KFIR_D_RECORD_START);
}

static void Kfir_Stop(struct kfir_dev *kfir)
{
	kfir->FlStopVideo = 1; 
	/*
	  KfirRegRead(KFIR_ADR_RECORD);
	  KfirRegWrite(KFIR_ADR_RECORD, KFIR_D_RECORD_STOP);
	  KfirRegRead(KFIR_ADR_RECORD);
	*/
}

static void InitParams(struct kfir_dev *kfir)
{
	VT_KFIR_CParams *P=&kfir->Params;
	DWORD video_offset, audio_offset;
	const WORD* RateArray;
	
	P->AspectRatio			= VT_KFIR_ASPECT_4_3;
	P->IntraDCPrecision		= mpeg1mode ? VT_KFIR_PREC_8_BITS : VT_KFIR_PREC_9_BITS;
	P->ColorPrimaries		= VT_KFIR_COLOR_ITUR_624_4BG;
	P->TransferCharacteristics	= VT_KFIR_TRANS_ITUR_624_4BG;
	P->MatrixCoefficients           = VT_KFIR_MATRIX_ITUR_624_4BG;
        P->FramePredictionDCT_I         = VT_KFIR_FRAME; 
        P->FramePredictionDCT_B         = VT_KFIR_FRAME;  
        P->FramePredictionDCT_P         = VT_KFIR_FRAME;  
        P->IntraVLCFormat_I             = mpeg1mode? VT_KFIR_B14 : VT_KFIR_B15; 
	P->IntraVLCFormat_B		= VT_KFIR_B14;
	P->IntraVLCFormat_P		= VT_KFIR_B14;
	P->AlternateScan_I		= mpeg1mode ? VT_KFIR_Zigzag : VT_KFIR_Alternate;
	P->AlternateScan_B		= mpeg1mode ? VT_KFIR_Zigzag : VT_KFIR_Alternate;
	P->AlternateScan_P		= mpeg1mode ? VT_KFIR_Zigzag : VT_KFIR_Alternate;
	P->ProgressiveSequence	        = mpeg1mode ? VT_KFIR_ON : VT_KFIR_OFF;
	P->ProgressiveFrame		= mpeg1mode ? VT_KFIR_ON : VT_KFIR_OFF;
	P->SliceBypass		        = VT_KFIR_OFF;
	P->VIB_FilterBypass		= VT_KFIR_ON;
	P->VIB_ProgressiveSource	= VT_KFIR_OFF;
	P->Video_Source		        = vidinput; 
	P->Video_Format		        = vidoutmode;
	P->AudioMode		        = VT_KFIR_STEREO;
	P->AudioFormat		        = VT_KFIR_LAYER_II;
	P->AudioFreq                    = sndfreq;
	P->AudioBitRate		        = 6;
	P->AudioCRC			= VT_KFIR_OFF;
	P->StreamType	                = streamtype;

	RateArray=ADSP_GetRates(kfir, kfir->Params.AudioMode, 
				kfir->Params.AudioFormat);
	P->TopField	 = 1;
	P->Bottom_Field	 = 1;
	P->Horiz	 = 0;
	P->VBFBufferSize = mpeg1mode ? 20 : 112; // 23 MPEG
	P->BitRate	 = vidrate;
	P->TransportRate = vidrate+RateArray[kfir->Params.AudioBitRate] * 1000;
	P->EachChunkPTS = VT_KFIR_OFF;
	P->VideoBufferSize=3;
	P->AudioBufferSize=3;
	P->VideoEventSize = 0;
	P->AudioEventSize = 1;

	P->VAD_LuminanceFactor      = 0;	/* 1 */
	P->VAD_LuminanceBandPass    = 1;	/* 1 */		/* use 2!!! */
	P->VAD_Prefilter            = 0;        /* use 1 if noise level is low - cause it's sharping !!! */
	P->VAD_LuminanceBrightness  = 0x80;
	P->VAD_LuminanceContrast    = 0x47;
	P->VAD_ChromaSaturation     = 64;
	P->VAD_HueControl           = 0x0;
	P->VAD_ChromLowPass         = 1;
	P->VAD_ChromCombFilterOff   = 1;
	P->VAD_ColorStandard        = 0;
	
	P->VAD_Dithering=0;
	P->VAD_ChromInterFilter=0;
	
	
	P->AudioStreamID		= 0xC0;
	P->VideoStreamID		= 0xE0;
	P->AudioProgramID		= 0x50;
	P->VideoProgramID		= 0xa0;
	P->ProgramMapID			= 0x400;
	P->SeqStart=1;
	P->FromFrame=0;
	P->ToFrame=100;
	P->PATVersion = 1;
	video_offset = (DWORD) (90000000UL/
				((P->Video_Format == VT_KFIR_PAL)?(2500):(2997)));
	audio_offset = (DWORD) (90000000UL/
				((P->Video_Format == VT_KFIR_PAL)?(2500):(2997)));
	P->InitialVideoPTS = video_offset;
	P->InitialVideoDTS = video_offset;
	P->InitialAudioPTS = audio_offset;
	P->InitialAudioDTS = audio_offset;
	P->ProgramNumber = 1;
	P->TransportStreamID = 0;
	
	// Kfir parameters initialization.
	if (vidoutmode==VT_KFIR_PAL)
		P->KF_FrameRate = KFIR_25;
	else
		P->KF_FrameRate = KFIR_30;
	
	P->KF_HorResolution = (vidsize&7);
	P->KF_VertResolution = ((vidsize>>4)&7);

	P->KF_VESAlignment = KFIR_VES_32;
	P->KF_RateControl = KFIR_RATE_CBR; //mpeg1mode ? KFIR_RATE_CBR : KFIR_RATE_VBR;
	P->KF_QuantScale = 112;
	P->KF_GOP_N = 15;
	P->KF_GOP_M = KFIR_IPBB_PICT;
	P->KF_SIF_PictureBottomField = 0;
	P->KF_LowDelay = 0;
	P->KF_LowSearchRange = 0;
	// P->KF_Auto_3_2_PullDown = 1;
	P->KF_Auto_3_2_PullDown = P->KF_GOP_M == KFIR_IPBB_PICT ? 1 : 0;
	P->KF_Auto_3_2_PullDown = 0;
	P->KF_Threshold = 2;
	
	// Kfir Filtering parameters.
	P->KF_HorLuminFilter = 1;
	P->KF_HorLuminFiltC0 = 38;
	P->KF_HorLuminFiltC1 = -11;
	P->KF_HorLuminFiltC2 = 113;
	P->KF_HorLuminFiltC3 = 924;
	P->KF_HorLuminFiltC4 = 53;
	P->KF_HorLuminFiltC5 = -73;
	P->KF_HorLuminFiltC6 = -22;
	P->KF_HorChromaFilter = 1;
	P->KF_HorChromaFiltC0 = 38;
	P->KF_HorChromaFiltC1 = -11;
	P->KF_HorChromaFiltC2 = 113;
	P->KF_HorChromaFiltC3 = 924;
	P->KF_HorChromaFiltC4 = 53;
	P->KF_HorChromaFiltC5 = -73;
	P->KF_HorChromaFiltC6 = -22;
	P->KF_VertChromaTopFilter = 1;
	P->KF_VertChromaTopFiltCenter = 0;
	P->KF_VertChromaTopFiltC0 = -40;
	P->KF_VertChromaTopFiltC1 = 216;
	P->KF_VertChromaTopFiltC2 = 480;
	P->KF_VertChromaTopFiltC3 = 376;
	P->KF_VertChromaTopFiltC4 = 56;
	P->KF_VertChromaTopFiltC5 = -64;
	P->KF_VertChromaBottomFilter = 1;
	P->KF_VertChromaBottomFiltCenter = 1;
	P->KF_VertChromaBottomFiltC0 = -64;
	P->KF_VertChromaBottomFiltC1 = 56;
	P->KF_VertChromaBottomFiltC2 = 376;
	P->KF_VertChromaBottomFiltC3 = 480;
	P->KF_VertChromaBottomFiltC4 = 216;
	P->KF_VertChromaBottomFiltC5 = -40;
	
	P->KF_Copyright = 0;
	P->KF_Original = 1;
	
	P->KF_TopFiledFirst = 0;  
	P->KF_FrameSkipRatio = 1;
	P->KF_3_2_PullDown = KFIR_AUTO;
	
	P->KF_Mpeg_I = mpeg1mode;
}


static int Kfir_DownLoadParams(struct kfir_dev *kfir)
{
	DWORD MulFrameRate;
	DWORD FrameRate;
	DWORD r;
	DWORD mbWidth;
	DWORD mbHeight;
	DWORD Data;
	DWORD MulBitRate = kfir->Params.BitRate * kfir->Params.KF_FrameSkipRatio;
	int i;
        VT_KFIR_CParams *Params=&kfir->Params;
	
	dprintk("kfir: Download parameters...");

	kfir->Par[KFIR_ADR_ASPECT_RATIO / 4] = Params->AspectRatio + 1; 
	kfir->Par[KFIR_ADR_FRAME_RATE_CODE / 4] = Params->KF_FrameRate + 1;
	kfir->Par[KFIR_ADR_COLOR_PRIM / 4] = Params->ColorPrimaries + 1;
	kfir->Par[KFIR_ADR_TRANSFER_CHAR/4] = Params->TransferCharacteristics + 1;
	kfir->Par[KFIR_ADR_MATRIX_COEFF/4] = Params->MatrixCoefficients + 1;
	
	Data = MulBitRate/400;
	kfir->Par[KFIR_ADR_BIT_RATE_0/4] = (WORD)(Data & 0xff);
	kfir->Par[KFIR_ADR_BIT_RATE_1/4] = (WORD)((Data & 0xff00) >> 8);
	kfir->Par[KFIR_ADR_BIT_RATE_2/4] = (WORD)((Data & 0xff0000) >> 16);
	
	kfir->Par[KFIR_ADR_VBF_BUFFER_SIZE/4] = Params->VBFBufferSize;
	kfir->Par[KFIR_ADR_PROGRESSIVE_SEQ/4] = Params->ProgressiveSequence;
	kfir->Par[KFIR_ADR_PROGRESSIVE_FRAME/4] = Params->ProgressiveFrame;

	kfir->Par[KFIR_ADR_INTRA_DC_PRECISION/4] = Params->IntraDCPrecision;
	kfir->Par[KFIR_ADR_INTRA_VLC_FORMAT_I/4] = Params->IntraVLCFormat_I;
	kfir->Par[KFIR_ADR_INTRA_VLC_FORMAT_P/4] = Params->IntraVLCFormat_P;
	kfir->Par[KFIR_ADR_INTRA_VLC_FORMAT_B/4] = Params->IntraVLCFormat_B;
	
	kfir->Par[KFIR_ADR_ALT_SCAN_I/4] = Params->AlternateScan_I;
	kfir->Par[KFIR_ADR_ALT_SCAN_P/4] = Params->AlternateScan_P ;
	kfir->Par[KFIR_ADR_ALT_SCAN_B/4] = Params->AlternateScan_B;
	
	kfir->Par[KFIR_ADR_FRAME_PRED_FRAME_DCT_I/4] = Params->FramePredictionDCT_I;
	kfir->Par[KFIR_ADR_FRAME_PRED_FRAME_DCT_P/4] = Params->FramePredictionDCT_P;
	kfir->Par[KFIR_ADR_FRAME_PRED_FRAME_DCT_B/4] = Params->FramePredictionDCT_B;

	kfir->Par[KFIR_ADR_AVG_ACT/4] = KFIR_D_AVG_ACT;
	kfir->Par[KFIR_ADR_X_I/4] = (WORD)(RMUL(160l,MulBitRate,115l) >> 17);
	kfir->Par[KFIR_ADR_X_P/4] = (WORD)(RMUL(60l,MulBitRate,115l) >> 17);
	kfir->Par[KFIR_ADR_X_B/4] = (WORD)(RMUL(42l,MulBitRate,115l) >> 17);
	
	switch (Params->KF_HorResolution) {
	case KFIR_H_160:
		mbWidth = 160 / 16;
		break;
	case KFIR_H_320:
		mbWidth = 320 / 16;
		break;
	case KFIR_H_352:
		mbWidth = 352 / 16;
		break;
	case KFIR_H_480:
		mbWidth = 480 / 16;
		break;
	case KFIR_H_544:
		mbWidth = 544 / 16;
		break;
	case KFIR_H_640:
		mbWidth = 640 / 16;
		break;
	default:
	case KFIR_H_FD1:
		mbWidth = 704 / 16;
		break;
	}

	if (Params->Video_Format) {  // PAL
		switch (Params->KF_VertResolution) {
		case KFIR_V_144:
			mbHeight = 144 / 16;
			break;
		case KFIR_V_288:
			mbHeight = 288 / 16;
			break;
		default:
		case KFIR_V_576:
			mbHeight = 576 / 16;
			break;    
		}
	}
	else {  // NTSC
		switch (Params->KF_VertResolution) {
		case KFIR_V_112:
			mbHeight = 112 / 16;
			break;
		case KFIR_V_240:
			mbHeight = 240 / 16;
			break;
		default:
		case KFIR_V_480:
			mbHeight = 480 / 16;
			break;
		}
	}

	switch (Params->KF_FrameRate) {
	case KFIR_23_976:
		MulFrameRate = 23976;
		break;
	case KFIR_24:
		MulFrameRate = 24000;
		break;
	default:
	case KFIR_25:
		MulFrameRate = 25000;
		break;
	case KFIR_29_97:
		MulFrameRate = 29970;
		break;
	case KFIR_30:
		MulFrameRate = 30000;
		break;
	}
	FrameRate=((double)MulFrameRate)/1000.0;
	r = (((4000 * MulBitRate) / MulFrameRate) + 1)/2;
	
	kfir->Par[KFIR_ADR_D0_I_0/4] = (WORD)((RMUL(10l,r,31l)>>2) & 0xffff);
	kfir->Par[KFIR_ADR_D0_P_0/4] = (WORD)((RMUL(10l,r,31l)>>2) & 0xffff);
	kfir->Par[KFIR_ADR_D0_B_0/4] = (WORD)((RMUL(14l,r,31l)>>2) & 0xffff);

	kfir->Par[KFIR_ADR_D0_I_1/4] = (WORD)((RMUL(10l,r,31l)>>18)& 0xffff);
	kfir->Par[KFIR_ADR_D0_P_1/4] = (WORD)((RMUL(10l,r,31l)>>18)& 0xffff);
	kfir->Par[KFIR_ADR_D0_B_1/4] = (WORD)((RMUL(14l,r,31l)>>18)& 0xffff);
	
	kfir->Par[KFIR_ADR_INTRAD/4] = 9*64;

	kfir->Par[KFIR_ADR_TMIN/4] = (WORD)(((1000*MulBitRate) / MulFrameRate)/4/4);
	
	kfir->Par[KFIR_ADR_R2/4] = (WORD)(RMUL(31L, (1<<22), r));
	kfir->Par[KFIR_ADR_MB_WIDTH/4] = (WORD)mbWidth;
	kfir->Par[KFIR_ADR_RMBW/4] = (WORD)(0x10000/mbWidth);
	kfir->Par[KFIR_ADR_RMB/4] = (WORD)(0x10000/(mbWidth*mbHeight));
	kfir->Par[KFIR_ADR_GETB/4] = (WORD)(90000L/(MulBitRate >> 13));

	kfir->Par[KFIR_ADR_C1/4] = (WORD)(256 * 1.4);
	kfir->Par[KFIR_ADR_C2/4] = (WORD)((1<<16)/1.4);

	kfir->Par[KFIR_ADR_STD_PICT_DEL/4] = (WORD)(90000000UL/MulFrameRate);
	kfir->Par[KFIR_ADR_FIRST_PICT_DEL/4] = 
		(WORD) ((((((double)Params->VBFBufferSize)*16384.0*2.0)/8.0)*90000.0)/((double)MulBitRate));


	Data=RMUL(MulBitRate*2, 
		  (Params->KF_GOP_N-Params->KF_GOP_M+1-(Params->KF_GOP_M==0)),
		  MulFrameRate/125);
	kfir->Par[KFIR_ADR_GOPR1_1/4] = (Data>>16);
	kfir->Par[KFIR_ADR_GOPR1_0/4] = Data;

	Data=RMUL(MulBitRate*2, 
		  Params->KF_GOP_N,
		  MulFrameRate/125);
	kfir->Par[KFIR_ADR_GOPR2_1/4] = (Data>>16);
	kfir->Par[KFIR_ADR_GOPR2_0/4] = Data;

	kfir->Par[KFIR_ADR_SLICE_CONTROL/4] = Params->SliceBypass;
	kfir->Par[KFIR_ADR_VIDEO_FORMAT/4] = Params->Video_Format;
	kfir->Par[KFIR_ADR_INTERLACED_SOURCE/4] = Params->VIB_ProgressiveSource;

	if (Params->KF_GOP_M < 0 || Params->KF_GOP_M > 3)
		return -1;
	kfir->Par[KFIR_ADR_M/4] = (WORD)Params->KF_GOP_M;

	if (Params->KF_GOP_M == 0)
		Data = Params->KF_GOP_N;
	else 
		Data = (Params->KF_GOP_N / Params->KF_GOP_M);
	kfir->Par[KFIR_ADR_MULTIPLE/4] = (WORD)Data;
	
	kfir->Par[KFIR_ADR_TOTAL_MBS/4] = (WORD)(mbWidth*mbHeight);
	kfir->Par[KFIR_ADR_MB_LINES/4] = (WORD)mbHeight;
	kfir->Par[KFIR_ADR_RANDOM_SCAN/4] = (WORD)Params->SeqStart;
	kfir->Par[KFIR_ADR_VBR_ENABLE/4] = (WORD)Params->KF_RateControl;

	if (Params->Video_Format == VT_KFIR_PAL) {
		kfir->Par[KFIR_ADR_TOP_FIRST/4] = 0;
	}
	else {
		kfir->Par[KFIR_ADR_TOP_FIRST/4] = (WORD)Params->KF_TopFiledFirst;
	}
	
	kfir->Par[KFIR_ADR_LOW_SEARCH_RANGE/4] = (WORD)Params->KF_LowSearchRange;
	kfir->Par[KFIR_ADR_CONST_MQUANT/4] = (WORD)(Params->KF_QuantScale & 0x3FF);
	kfir->Par[KFIR_ADR_LOW_LATENCY/4] = (WORD)Params->KF_LowDelay;
	kfir->Par[KFIR_ADR_AUTO_PULLDOWN/4] = (WORD)Params->KF_Auto_3_2_PullDown;
	kfir->Par[KFIR_ADR_BITSTREAM_WIDTH/4] = (WORD)KFIR_D_BITSTREAM_WIDTH_32;
	kfir->Par[KFIR_ADR_VOB_BASE_ADDR/4] = VOB_BASE_ADDR;
	kfir->Par[KFIR_ADR_USE_SDRAM_FIFO/4] = KFIR_D_USE_SDRAM_FIFO_ON;

	switch (Params->KF_HorResolution) {
	case KFIR_H_160:
		Data = KFIR_D_VIB_DEC_PATTERN_160;
		break;
	case KFIR_H_320:
		Data = KFIR_D_VIB_DEC_PATTERN_320;
		break;
	case KFIR_H_352:
		Data = KFIR_D_VIB_DEC_PATTERN_352;
		break;
	case KFIR_H_480:
		Data = KFIR_D_VIB_DEC_PATTERN_480;
		break;
	case KFIR_H_544:
		Data = KFIR_D_VIB_DEC_PATTERN_544;
		break;
	case KFIR_H_640:
		Data = KFIR_D_VIB_DEC_PATTERN_640;
		break;
	case KFIR_H_FD1:
		Data = KFIR_D_VIB_DEC_PATTERN_FD1;
		break;
	default:
		break;
	}
	kfir->Par[KFIR_ADR_VIB_DEC_PATTERN/4] = (WORD)Data;
	
	switch (Params->KF_HorResolution) {
	case KFIR_H_160:
		Data = KFIR_ADR_VIB_DEC_SIZE_160;
		break;
	case KFIR_H_320:
		Data = KFIR_ADR_VIB_DEC_SIZE_320;
		break;
	case KFIR_H_352:
		Data = KFIR_ADR_VIB_DEC_SIZE_352;
		break;
	case KFIR_H_480:
		Data = KFIR_ADR_VIB_DEC_SIZE_480;
		break;
	case KFIR_H_544:
		Data = KFIR_ADR_VIB_DEC_SIZE_544;
		break;
	case KFIR_H_640:
		Data = KFIR_ADR_VIB_DEC_SIZE_640;
		break;
	case KFIR_H_FD1:
		Data = KFIR_ADR_VIB_DEC_SIZE_FD1;
		break;
	default:
		break;
	}
	kfir->Par[KFIR_ADR_VIB_DEC_SIZE/4] = (WORD)Data;

	Data = (Params->KF_HorLuminFilter            ? 0x0001 : 0) |
		(Params->KF_HorChromaFilter           ? 0x0002 : 0) |
		(Params->Video_Source == VT_KFIR_FILE  ? 0x0040 : 0) |
		(Params->KF_TopFiledFirst             ? 0x0080 : 0) |
		(Params->Video_Format == VT_KFIR_NTSC  ? 0      : 0x0100);
	if (Params->KF_VertChromaTopFilter)
		Data |= Params->KF_VertChromaTopFiltCenter ? 0x000C : 0x0008;
	if (Params->KF_VertChromaBottomFilter)
		Data |= Params->KF_VertChromaBottomFiltCenter ? 0x0030 : 0x0020;
	Data |= 0x0200;
	kfir->Par[KFIR_ADR_VIB_VIDEO_PAR/4] = (WORD)Data;

	Data = (Params->Horiz & KFIR_D_VIB_HORI_TOP_MASK);
	kfir->Par[KFIR_ADR_VIB_HORI_TOP_BOT/4] = (WORD)((Data << 4) + Data);

	Data = ((Params->TopField << 5) & KFIR_D_VIB_VERT_TOP_MASK) |
		(Params->Bottom_Field & KFIR_D_VIB_VERT_BOT_MASK);
	kfir->Par[KFIR_ADR_VIB_VERT_TOP_BOT/4] = (WORD)Data;


	kfir->Par[KFIR_ADR_VIB_FILTERC_YH_0/4] = Params->KF_HorLuminFiltC0;
	kfir->Par[KFIR_ADR_VIB_FILTERC_YH_1/4] = Params->KF_HorLuminFiltC1;
	kfir->Par[KFIR_ADR_VIB_FILTERC_YH_2/4] = Params->KF_HorLuminFiltC2;
	kfir->Par[KFIR_ADR_VIB_FILTERC_YH_3/4] = Params->KF_HorLuminFiltC3;
	kfir->Par[KFIR_ADR_VIB_FILTERC_YH_4/4] = Params->KF_HorLuminFiltC4;
	kfir->Par[KFIR_ADR_VIB_FILTERC_YH_5/4] = Params->KF_HorLuminFiltC5;
	kfir->Par[KFIR_ADR_VIB_FILTERC_YH_6/4] = Params->KF_HorLuminFiltC6;
	kfir->Par[KFIR_ADR_VIB_FILTERC_CH_0/4] = Params->KF_HorChromaFiltC0;
	kfir->Par[KFIR_ADR_VIB_FILTERC_CH_1/4] = Params->KF_HorChromaFiltC1;
	kfir->Par[KFIR_ADR_VIB_FILTERC_CH_2/4] = Params->KF_HorChromaFiltC2;
	kfir->Par[KFIR_ADR_VIB_FILTERC_CH_3/4] = Params->KF_HorChromaFiltC3;
	kfir->Par[KFIR_ADR_VIB_FILTERC_CH_4/4] = Params->KF_HorChromaFiltC4;
	kfir->Par[KFIR_ADR_VIB_FILTERC_CH_5/4] = Params->KF_HorChromaFiltC5;
	kfir->Par[KFIR_ADR_VIB_FILTERC_CH_6/4] = Params->KF_HorChromaFiltC6;
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV0_0/4] = Params->KF_VertChromaTopFiltC0;
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV0_1/4] = Params->KF_VertChromaTopFiltC1;
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV0_2/4] = Params->KF_VertChromaTopFiltC2;
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV0_3/4] = Params->KF_VertChromaTopFiltC3;
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV0_4/4] = Params->KF_VertChromaTopFiltC4;
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV0_5/4] = Params->KF_VertChromaTopFiltC5;
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV1_0/4] = Params->KF_VertChromaBottomFiltC0;
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV1_1/4] = Params->KF_VertChromaBottomFiltC1;
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV1_2/4] = Params->KF_VertChromaBottomFiltC2;
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV1_3/4] = Params->KF_VertChromaBottomFiltC3;
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV1_4/4] = Params->KF_VertChromaBottomFiltC4;
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV1_5/4] = Params->KF_VertChromaBottomFiltC5;
	
	Data = ((0x1<<Params->KF_Threshold) - 1)*mbWidth*mbHeight*256/2;
	kfir->Par[KFIR_ADR_TELE_THRESHOLD_0/4] = (WORD)(Data & 0xffff);
	kfir->Par[KFIR_ADR_TELE_THRESHOLD_1/4] = (WORD)((Data >> 16)&0xffff);
	kfir->Par[KFIR_ADR_DMA_BSM_BURST/4] = 200;// 64;
	kfir->Par[KFIR_ADR_DMA_BSM_BURST_END/4] = 10;// 60;
	kfir->Par[KFIR_ADR_DMA_VIB_BURST/4] = 200;
	kfir->Par[KFIR_ADR_DMA_VIB_BURST_END/4] = 10;
	kfir->Par[KFIR_ADR_FILE_MODE/4] = Params->Video_Source == VT_KFIR_FILE ? 1 : 0;
	{
		DWORD gopr = RMUL(MulBitRate*2, Params->KF_GOP_N, MulFrameRate/125);
		//(DWORD)((Params->KF_GOP_N*MulBitRate)/FrameRate)/4;
		DWORD vtmp1;
		DWORD init_T;
		
		if(Params->KF_GOP_M == 0)
			vtmp1 = 1;
		else
			vtmp1 = (5 + 
				 2 *(Params->KF_GOP_N/Params->KF_GOP_M - 1) + 
				 (Params->KF_GOP_N -
				  (Params->KF_GOP_N/Params->KF_GOP_M - 1)-1));


		init_T = (DWORD)(gopr/vtmp1);
		
		kfir->Par[101*4/4] = (WORD)((init_T >> 16) & 0xFFFF);
		kfir->Par[102*4/4] = (WORD)(init_T & 0xFFFF);
	}

	kfir->Par[KFIR_ADR_MPEG_I_NEG/4] = Params->KF_Mpeg_I ? 0 : 1;
	kfir->Par[KFIR_ADR_FRAME_SKIP_RATIO/4] = Params->KF_FrameSkipRatio;

	for (i=0; i<KFIR_MAX_PARAMS; i++) 
		KfirRegWrite(i*4, kfir->Par[i]);

	printk("kfir: done\n");
	return 0;
}

static int Kfir_CheckParams(struct kfir_dev *kfir)
{
	int eRet = 0;
	int i;
	DWORD Data[128];
	
	for (i=0; i<128; i++) {
		Data[i] = 0xffff & KfirRegRead(i*4);
		if (Data[i] != kfir->Par[i])
			printk("kfir: CheckParam: Par[%d] = %x != Read Data[%d]=%x\n",
			       i, kfir->Par[i], i, Data[i]);
	}
	
  	return eRet;
}

static int KfirGeneric_StartNDownLoadParams(struct kfir_dev *kfir)
{
	int ret;
	int Step;
	DWORD Data;

	AlteraRegWrite(ALT_ADR_CLOCK_PADI_DEF, 0);
	
	KfirRegWrite(KFIR_200, 7 << 8);
	KfirRegWrite(KFIR_201, 0);

	KfirRegWrite(KFIR_200, (7<<8) + 5);
	Data = KfirRegRead(KFIR_201);

	Step = 0;
	do {
		Sleep(100);
		if (Step++ == 10)
			return -1;
		Data = (0xffff&KfirRegRead(KFIR_ADR_EREADY));
	} while (Data != KFIR_D_EREADY_READY);
	
	if ((ret = Kfir_DownLoadParams(kfir))<0) {
		return ret;
	}
	if ((ret = Kfir_CheckParams(kfir))<0) {
		return ret;
	}
	if ((ret = Kfir_LoadAudioTable(kfir, KF_AUDIO_MEM_BASE)) <0) {
		return ret;
	}
	return 0;
}

static void Kfir_Reset(struct kfir_dev *kfir)
{
	if (kfir->kfirreset)
		return;

	printk("kfir: Reset Kfir and PLL ... ");

	AlteraRegWrite(ALT_ADR_CLOCK_PADI_DEF, 0);

	waltl(ALT_ADR_RESET_REG, 0);
	mdelay(10);
	waltl(ALT_ADR_RESET_REG, ALT_D_RESET_REG_KFIR_PLL_REV_5E);
	mdelay(10);
	waltl(ALT_ADR_RESET_REG, 0);
	mdelay(10);
	waltl(ALT_ADR_RESET_REG, ALT_D_RESET_REG_KFIR_REV_5E);
	mdelay(3);

        kfir->kfirreset=1;

	printk("done\n");
}


/* ----------------------------------------------------------------------- */
/* I2C functions                                                           */

static void kfir_bit_setscl(void *data, int state)
{
	struct kfir_dev *kfir = (struct kfir_dev *) data;

	if (state)
		kfir->i2c |= 0x01;
	else
		kfir->i2c &= ~0x01;
        waltl(ALT_ADR_I2C, kfir->i2c); 
        udelay(150);
}

static void kfir_bit_setsda(void *data, int state)
{
	struct kfir_dev *kfir = (struct kfir_dev *) data;

	if (state)
		kfir->i2c |= 0x02;
	else
		kfir->i2c &= ~0x02;
	waltl(ALT_ADR_I2C, kfir->i2c); 
	udelay(150);
}

static int kfir_bit_getscl(void *data)
{
	struct kfir_dev *kfir = (struct kfir_dev *) data;
	int state;
	
	state = raltl(ALT_ADR_I2C) & 0x01 ? 1 : 0;
	return state;
}

static int kfir_bit_getsda(void *data)
{
	struct kfir_dev *kfir = (struct kfir_dev *) data;
	int state;

	state = raltl(ALT_ADR_I2C) & 0x02 ? 1 : 0;
	return state;
}

static void kfir_inc_use(struct i2c_adapter *adap)
{
	MOD_INC_USE_COUNT;
}

static void kfir_dec_use(struct i2c_adapter *adap)
{
	MOD_DEC_USE_COUNT;
}

static int attach_inform(struct i2c_client *client)
{
        struct kfir_dev *kfir = (struct kfir_dev *) client->adapter->data;
	
	if (client->driver->id==I2C_DRIVERID_SAA7113) {
		if (!kfir->decoder) 
			kfir->decoder=client;
		printk("kfir: saa7113 decoder attached\n");
	}
	
	printk("kfir: i2c attach [%s]\n", client->name);
        return 0;
}

static int detach_inform(struct i2c_client *client)
{
        //struct kfir_dev *kfir = (struct kfir_dev *) client->adapter->data;
	
	printk("kfir: i2c detach [%s]\n", client->name);
        return  0;
}

static struct i2c_algo_bit_data i2c_algo_template = {
	NULL, 
	kfir_bit_setsda,
	kfir_bit_setscl,
	kfir_bit_getsda,
	kfir_bit_getscl,
	10, 10, 100,
};

static struct i2c_adapter i2c_adap_template = {
	"kfir",
	I2C_HW_B_BT848,
	NULL,
	NULL,
	kfir_inc_use,
	kfir_dec_use,
	attach_inform,
	detach_inform,
	NULL,
};

static struct i2c_client i2c_client_template = {
        "kfir internal",
        -1,
        0,
        0,
        NULL,
        NULL
};

static int init_i2c(struct kfir_dev *kfir)
{
	/* i2c bit_adapter */
	memcpy(&kfir->i2c_adap, &i2c_adap_template, sizeof(struct i2c_adapter));
	memcpy(&kfir->i2c_algo, &i2c_algo_template, sizeof(struct i2c_algo_bit_data));
	memcpy(&kfir->i2c_client, &i2c_client_template, sizeof(struct i2c_client));

	sprintf(kfir->i2c_adap.name+strlen(kfir->i2c_adap.name), " #%d", num_kfir-1);
        kfir->i2c_algo.data = kfir;
        kfir->i2c_adap.data = kfir;
        kfir->i2c_adap.algo_data = &kfir->i2c_algo;
        kfir->i2c_client.adapter = &kfir->i2c_adap;

	kfir_bit_setscl(kfir,1);
	kfir_bit_setsda(kfir,1);

	kfir->i2c_ok = i2c_bit_add_bus(&kfir->i2c_adap);
	return kfir->i2c_ok;
}

static void scani2cbus(struct kfir_dev *kfir)
{
        struct i2c_adapter *adap=&kfir->i2c_adap;
	int i;
        struct i2c_msg msgs[1];
	u8 mm1;

	for (i=0; i<0x7f; i++) {
		msgs[0].flags=0;
		msgs[0].addr=i;
		msgs[0].len=1;
		msgs[0].buf=&mm1;
		if (i2c_transfer(adap, msgs, 1)==1)
			printk("kfir: Found i2c device at: 0x%02x\n", 2*i);
	}
}

static int decoder_command(struct kfir_dev *kfir, unsigned int cmd, void *arg)
{
        if (!kfir->decoder)
                return -1;
        return kfir->decoder->driver->command(kfir->decoder, cmd, arg);
}


/****************************************************************************/
/* Altera functions                                                         */
/****************************************************************************/

static int AlteraWriteMicro(struct kfir_dev *kfir, u8 *micro, u32 microlen)
{
	u32 nConfig, EEPROMReg, n;
	u32 Data, Data1, GPInput;
  	u8 bVal;
	int i;
  
	// get register with nConfig.
	pci_read_config_dword(kfir->pdev, PLX9054_CONF_HOT_SWAP, &nConfig);

	// get register with DCLK and Data.
	EEPROMReg=rplxl(PLX9054_CNTRL_OFFSET);

	// Set nConfig = 0;
	Data = nConfig & ~HOT_SWAP_LED;
	pci_write_config_dword(kfir->pdev, PLX9054_CONF_HOT_SWAP, Data);
	// Set DCLK = 0;
	Data1 = EEPROMReg & ~SERIAL_EEPROM_CLK;
	wplxl(PLX9054_CNTRL_OFFSET, Data1);
	
	udelay(2);

	// Set nConfig = 1
	Data = nConfig | HOT_SWAP_LED;
	pci_write_config_dword(kfir->pdev, PLX9054_CONF_HOT_SWAP, Data);
	
	udelay(10);
	
	GPInput=rplxl(PLX9054_CNTRL_OFFSET);
	
	// begin download code loop
	n = 0;			// initial offset
	Data1 = EEPROMReg;
	while(n < microlen) {
		bVal = micro[n++];
		for (i = 0 ; i < 8 ; i++) {
			// set DCLK = 0
			Data1 = Data1 & ~SERIAL_EEPROM_CLK;
			wplxl(PLX9054_CNTRL_OFFSET, Data1);
			udelay(5);
			
			// Set Data bit.
			if (bVal & 0x01)
				Data1 |= SERIAL_EEPROM_DATA;	// data high
			else
				Data1 &= ~SERIAL_EEPROM_DATA;  //  data Low
			// Output Data with clock low
			wplxl(PLX9054_CNTRL_OFFSET, Data1);
			udelay(5);

			// set DCLK = 1
			Data1 |= SERIAL_EEPROM_CLK;
			wplxl(PLX9054_CNTRL_OFFSET, Data1);
			udelay(5);

			bVal >>= 1;
		}
	}
	ddelay(10); mdelay(100);
	Data1 &= ~SERIAL_EEPROM_CLK;
	wplxl(PLX9054_CNTRL_OFFSET, Data1);
  
	udelay(2);

	// begin Altera init loop
	for (i = 0 ; i < 20 ; i++) {
		// Set DCLK = 0;
		Data1 &= ~SERIAL_EEPROM_CLK;
		wplxl(PLX9054_CNTRL_OFFSET, Data1);
		
		udelay(5);
		
		// set DCLK = 1
		Data1 |= SERIAL_EEPROM_CLK;
		wplxl(PLX9054_CNTRL_OFFSET, Data1);
		
		udelay(5);
	}
	// Set DCLK = 0;
	Data1 = EEPROMReg & ~SERIAL_EEPROM_CLK;
	Data1 &= ~SERIAL_EEPROM_DATA;
	wplxl(PLX9054_CNTRL_OFFSET, Data1);
	
	GPInput=rplxl(PLX9054_CNTRL_OFFSET);

	return 0;
}

static int AlteraReset(struct kfir_dev *kfir)
{
	u32 LAS0RR_0, LAS0BA_0, MARBR_0, LBRD0_0, DMRR_0, DMLBAM_0;
	u32 DMLBA1_0, DMPBAM_0, LAS1RR_0, LAS1BA_0, LBRD1_0, INTCSR_0, LROMCBLE_0;
	u32 Data;

	printk("kfir: Reset Altera\n");
	ddelay(10); 

	// Check ConfDone
	Data=rplxl(PLX9054_CNTRL_OFFSET);
	if (!(Data & GP_INPUT)) {
		return -1;
	}

	// Reset Altera.
	// Remember all parameters
	LAS0RR_0=rplxl(PLX9054_LAS0RR_OFFSET);
	LAS0BA_0=rplxl(PLX9054_LB0_BASE_ADR);
	MARBR_0=rplxl(PLX9054_MARBR_OFFSET);
	LBRD0_0=rplxl(PLX9054_LBRD0_OFFSET);
	DMRR_0=rplxl(PLX9054_DMRR_OFFSET);
	DMLBAM_0=rplxl(PLX9054_DMLBAM_OFFSET);
	DMLBA1_0=rplxl(PLX9054_DMLBA1_OFFSET);
	DMPBAM_0=rplxl(PLX9054_DMPBAM_OFFSET);
	LAS1RR_0=rplxl(PLX9054_LAS1RR_OFFSET);
	LAS1BA_0=rplxl(PLX9054_LAS1BA_OFFSET);
	LBRD1_0=rplxl(PLX9054_LBRD1_OFFSET);
	INTCSR_0=rplxl(PLX9054_INT_CNTRL_STS);
	LROMCBLE_0=rplxl(PLX9054_BIGEND_LMISC_PROTAREA);
	
	Data=rplxl(PLX9054_CNTRL_OFFSET);
	Data |= 0x40000000;
	wplxl(PLX9054_CNTRL_OFFSET, Data);
	//Sleep(10);
	Data &= ~0x40000000;
	wplxl(PLX9054_CNTRL_OFFSET, Data);
	
	udelay(10000);
	
	wplxl(PLX9054_LAS0RR_OFFSET, LAS0RR_0);
	wplxl(PLX9054_LB0_BASE_ADR, LAS0BA_0);
	wplxl(PLX9054_MARBR_OFFSET, MARBR_0);
	wplxl(PLX9054_LBRD0_OFFSET, LBRD0_0);
	wplxl(PLX9054_DMRR_OFFSET, DMRR_0);
	wplxl(PLX9054_DMLBAM_OFFSET, DMLBAM_0);
	wplxl(PLX9054_DMLBA1_OFFSET, DMLBA1_0);
	wplxl(PLX9054_DMPBAM_OFFSET, DMPBAM_0);
	wplxl(PLX9054_LAS1RR_OFFSET, LAS1RR_0);
	wplxl(PLX9054_LAS1BA_OFFSET, LAS1BA_0);
	wplxl(PLX9054_LBRD1_OFFSET, LBRD1_0);
	wplxl(PLX9054_INT_CNTRL_STS, INTCSR_0);
	wplxl(PLX9054_BIGEND_LMISC_PROTAREA, LROMCBLE_0);
	
	// Mask interupts in Altera.
	waltl(ALT_ADR_INT_MASK, 0);
	// Clear the interrupts in Altera.
	Data = raltl(ALT_ADR_INT_REG_CLR);
	waltl(ALT_ADR_INT_REG_CLR, 0xFF);

	printk("kfir: done\n");
	ddelay(10);
	return 0;
}

static int AlteraLoadMicro(struct kfir_dev *kfir)
{
	u32 Data;

	u32 microlen;
	u8 *micro;
	int result=0;

	if (!loaded) {
		if (
#ifdef __KFIR_PACK__
		    !(microlen=firmread("micro/altera_keb_plx_5e.rbf", 
					(char**) &micro)) &&
#endif
		    !(microlen=firmread(KFIR_MICRO_PATH "altera_keb_plx_5e.rbf", 
					(char**) &micro)))
		return -1;

		printk("loaded microcode with length %d\n", microlen);
		ddelay(100); mdelay(1000);
		//Data=rplxl(PLX9054_CNTRL_OFFSET);
		
		result=AlteraWriteMicro(kfir, micro, microlen);

		vfree(micro);
		if (result<0)
			return result;
	}
	return AlteraReset(kfir);
}

static void start_dma0(struct kfir_dev *kfir)
{
	DWORD Data;

	// DMA 0 PCI start address
	wplxl(PLX9054_DMA0_PCI_ADDRESS, virt_to_bus(kfir->dmabuf));

	// DMA 0 LB start address
	Data = rplxl(PLX9054_LB0_BASE_ADR);
	Data &= LB_SPACE0_ADDRESS_MASK;
#ifdef ALTERA_BURST
	Data += 0x4000;
#else
	Data += KFIR_203 + 0x2000;
#endif
	wplxl(PLX9054_DMA0_LB_ADDRESS, Data);

	kfir->LastDMASize = MUX_DMA_LENGTH;
	wplxl(PLX9054_DMA0_SIZE, MUX_DMA_LENGTH);
	wplxl(PLX9054_INT_CNTRL_STS, rplxl(PLX9054_INT_CNTRL_STS)|DMA0_INT_ENABLE);
	wplxl(PLX9054_DMA_CMD_STS, 
	      rplxl(PLX9054_DMA_CMD_STS)|DMA0_ENABLE|DMA0_START);
}

static void 
KfProcessGenericModeInterrupt(struct kfir_dev *kfir)
{
	DWORD Data;
	
	do {
		ReadKfirStateGeneric(kfir);	
		
		if (kfir->FlStopVideo) { // Stop the chip.
			KfirRegWrite(KFIR_ADR_RECORD, KFIR_D_RECORD_STOP);
			Data = 0xffff&KfirRegRead(KFIR_ADR_RECORD);
			if (Data == KFIR_D_RECORD_STOP)
				kfir->FlStopVideo = 0; 
		}
		
		if (kfir->KfState.IntVideo & KDIR_D_INT_VIDEO) { // Kfir Video int.
			if (kfir->IntVideoCnt == 0) { // Start DMA transfer here.
				AlteraRegWrite(ALT_ADR_CLOCK_PADI_DEF, 1);
				start_dma0(kfir);
			}
			// Clear this bit.
			KfirRegWrite(KFIR_ADR_INT_VIDEO, 0);
			kfir->IntVideoCnt++;
		}
		
		// Kfir Audio int.    
		if (kfir->KfState.IntAudio & KDIR_D_INT_AUDIO) {   
			KfirRegWrite(KFIR_ADR_INT_AUDIO, 0);
			kfir->IntAudioCnt++;
		}
		
		// Kfir VIB interrupt
		if (kfir->KfState.IntDMA_VIB & KDIR_D_INT_DMA_VIB ) { 
			//KfirRegWrite(KFIR_ADR_INT_DMA_VIB, 0); 
			kfir->IntVibCnt++;
		}
		// Get int counter once more.
		kfir->KfState.IntCnt = KfirRegRead(KFIR_ADR_INTERRUPT_CNT);
	} while (kfir->KfState.OldIntCnt != kfir->KfState.IntCnt);
	
	return;
}

static void kfir_irq(int irq, void *dev_id, struct pt_regs * regs)
{
	struct kfir_dev *kfir= (struct kfir_dev *) dev_id;
	u32 plxstat;
	u32 altstat, altmask;
	u32 val;
	int count=0;

	while (1) {
		plxstat=rplxl(PLX9054_INT_CNTRL_STS);
		wplxl(PLX9054_INT_CNTRL_STS, plxstat);
		//dprintk("kfir: irq_stat=%08x\n", plxstat);

		if (!(plxstat&(LOCAL_INT_ACTIVE|DMA0_INT_ACTIVE|(1<<14))))
			return;

		/* Error on PCI bus ? */
		if (plxstat & (1<<14)) {
			pci_read_config_dword(kfir->pdev, PCI_COMMAND, &val);
			if (val&0xf8000000) 
				pci_write_config_dword(kfir->pdev, PCI_COMMAND, val);
		}
		
		/* Altera Interrupt ? */
		if (plxstat & LOCAL_INT_ACTIVE) {
			//dprintk("kfir: local int\n");
			altstat = raltl(ALT_ADR_INT_REG_CLR) & 0xff;
			       
			// audio interrupt
			if (altstat & ALT_D_INT_REQ_CLR_ADSP_INT) {
				printk("kfir: ADSP int\n");
				waltl(ALT_ADR_INT_REG_CLR, altstat);
				
				altmask = raltl(ALT_ADR_INT_MASK) & 0xFF;
				// Mask Audio Int.
				altmask &= ~ALT_D_INT_REQ_CLR_ADSP_INT;
				waltl(ALT_ADR_INT_MASK, altmask);
				
				if (kfir->Params.StreamType != VT_KFIR_VIDEO_ES) { 
					// nothing for 5E version
				}
				KfirRegWrite(KFIR_ADR_AUDIO_INT_ACK, 1);
			}
			
			// video interrupt
			if (altstat & ALT_D_INT_REQ_CLR_KFIR_INT) {
				waltl(ALT_ADR_INT_REG_CLR, altstat);
				if (kfir->mode == KFIR_GENERIC_MODE)
					KfProcessGenericModeInterrupt(kfir);
			}
		}
		
		if (kfir->mode == KFIR_GENERIC_MODE) { 
			// DMA0 interrupt?
			if (plxstat&DMA0_INT_ACTIVE) {
				DWORD Data;
				int free;

				// Clear DMA 0 interrupt.
				Data = rplxl(PLX9054_DMA_CMD_STS);
				Data |= DMA0_CLEAR_INT;
				Data &= ~DMA0_ENABLE;     // disable it.
				wplxl(PLX9054_DMA_CMD_STS, Data);
				
				if (kfir->KfState.KfirEReady == KFIR_D_EREADY_SEQ_END) {
				}
				if (kfir->KfState.KfirEReady == KFIR_D_EREADY_READY) {
					printk("kfir ready\n");
					//kfir->recording=0;
				}
				if (kfir->recording==1) {
					free=kfir->readp-kfir->writep;
					if (free<=0) free+=MPGBUFSIZE;
					if (free>MUX_DMA_LENGTH) {
						memcpy(kfir->mpgbuf+kfir->writep, 
						       kfir->dmabuf, MUX_DMA_LENGTH);
						kfir->writep+=MUX_DMA_LENGTH;
						kfir->writep&=(MPGBUFSIZE-1);
						wake_up(&kfir->wq);
						
					} // else we have an overflow!!
				}
				if (kfir->recording==2) {
					int i;
					u8 *ptr;
					u32 val;
					
					if (kfir->Params.StreamType == 
					     VT_KFIR_TRANSPORT) {
						kfir->recording=0;
						break;
					}

					for (i=0; i<MUX_DMA_LENGTH-3; i++) {
						ptr=kfir->dmabuf+i;
						val = ((ptr[0] << 24) | 
						       (ptr[1] << 16) |
						       (ptr[2] << 8) |
						       ptr[3]);
						
						if (val==0x000001b7) {
							dprintk("kfir: end code 1 found at %d\n",i);
							//kfir->recording=0;
							//break;
						}
						if (val==0x000001b9) {
							dprintk("kfir: end code 2 found at %d\n",i);
							kfir->recording=0;
							break;
						}

					}

				}
				if (kfir->recording) 
					start_dma0(kfir);

			}

		}
		if ((++count)>10) {
			printk("kfir: irq loop\n");
			wplxl(PLX9054_INT_CNTRL_STS, 0);
		}
	}
}


static int init_plx9054(struct kfir_dev *kfir)
{
	u32 val;
	int result=0;

	// Disable interrupts in PLX board (Global PCI Int, 
	// Global Local Int output, DMA0, DMA1, Doorbell to PCI, Doorbell to Local)
	val=rplxl(PLX9054_INT_CNTRL_STS);
	val &= ~(PCI_INT_ENABLE | LOCAL_INT_OUTPUT_ENABLE | 
		 DMA0_INT_ENABLE | DMA1_INT_ENABLE);
	// Disable PCI abort interrupt.
	val &= ~PCI_ABORT_INT_ENABLE;
	// Disable Master Retry indefinetely. // Obsolete
	val |= RETRY_ABORT_ENABLE;
	// Enable Local Int input
	val |= LOCAL_INT_INPUT_ENABLE;
	// Disable LSERR interrupts.
	val &= ~(LSERR_ENABLE);
	wplxl(PLX9054_INT_CNTRL_STS, val);
	
	// Enable DMA Ch0 Done and route DMA0 interrupt to PCI,
	// enable fast termination, hold Local Bus Address, enable demand mode.
	val = rplxl(PLX9054_DMA0_MODE);
	val |= (DMA0_DONE_ENABLE | DMA0_INT_TO_PCI);
#ifdef ALTERA_BURST
	val &= ~DMA0_HOLD_LOCAL_ADRESS;
#else
	val |= DMA0_HOLD_LOCAL_ADRESS;
#endif
	if (kfir->subsystem_device == 0x200)
		val |= DMA0_DEMAND_MODE;
	else
		val &= ~DMA0_DEMAND_MODE;
	// Must be Slow Termination for the revision 5E. This fixed the problem with
	// stucking during DMA transfer for the 16 bit width local bus. 28.11.1999 mgv
	val &= ~DMA0_FAST_TERMINATION;    // Slow termination
	// val |= DMA0_FAST_TERMINATION;    // Fast termination
	// Disable Local Burst and Bterm
	// val &= ~DMA0_BTERM_ENABLE;   // BTERM disable
	val |= DMA0_BTERM_ENABLE;       // BTERM enable
	// val &= ~DMA0_LOCAL_BURST_ENABLE;  // burst disable
#if PLX_BURST
	val |= DMA0_LOCAL_BURST_ENABLE;      // burst enable
#else
	val &= ~DMA0_LOCAL_BURST_ENABLE;
#endif

	val &= ~DMA0_LB_WIDTH_MASK;
	val |= DMA0_LB_WIDTH_16;
	wplxl(PLX9054_DMA0_MODE, val);
	
	// Set direction Local->PCI for the DMA channel 0.
	val = rplxl(PLX9054_DMA0_DESCR_PTR);
	val |= DMA0_LOCAL_TO_PCI_TRANSFER;
	// Disable Scater/Gather interrupt from each descriptor terminate counter.
	val &= ~DMA0_DESCRIPTOR_INT_ENABLE;
	wplxl(PLX9054_DMA0_DESCR_PTR, val);

	// Enable DMA Ch1 Done and route it to PCI.
	val = rplxl(PLX9054_DMA1_MODE);
	val |= (DMA1_DONE_ENABLE | DMA1_INT_TO_PCI);
	wplxl(PLX9054_DMA1_MODE, val);

	result = request_irq(kfir->pdev->irq, kfir_irq,
                             SA_SHIRQ | SA_INTERRUPT, "kfir", (void *) kfir);
	if (result<0)
		return result;

	// PCI-to-Local Space 0 (for Kfir and Altera).
	val = rplxl(PLX9054_LB0_BASE_ADR);
	wplxl(PLX9054_LB0_BASE_ADR, val);
	val = rplxl(PLX9054_LB0_BASE_ADR);
	
	// Disable prefetch and set maximim delay for the PCI bus.
	// This solved the problem with the write and immidiate read from the ADSP !!! (23.05.99)
	val = rplxl(PLX9054_LBRD0_OFFSET);
	val |= (SPACE0_PREFETCH_DISABLE | PCI_TARGET_RETRY_DELAY_MAX);
	wplxl(PLX9054_LBRD0_OFFSET, val);
	
	// Set read from local bus to be locked operation.
	val = rplxl(PLX9054_MARBR_OFFSET);
	val |= DIRECT_SLAVE_LOCK_ENABLE;
	// Clear the DMA channels priority
	val &= ~PRIOPITY_MASK;
	val |= PRIORITY_DMA_0;    // Set priority to DMA 0.
	wplxl(PLX9054_MARBR_OFFSET, val);
	
	// PCI-to-Local Space 1 (for Audio).
	rplxl(PLX9054_LB1_BASE_ADR);
	val &= ~LB_SPACE1_ADDRESS_MASK;
	val |= (SPACE1_ENABLED | KFIR_AUDIO_INPUT_BASE_LB_ADR);
	wplxl(PLX9054_LB1_BASE_ADR, val);
	
	// Swap bytes for the Direct master (MUX from Kfir and VIB to Kfir) and Direct Slave (Audio to Kfir)
	val = rplxl(PLX9054_BIGEND_LMISC_PROTAREA);
	val |= (BIGEND_DIRECT_MASTER | BIGEND_SPACE_1 | BIGEND_DMA0);
	wplxl(PLX9054_BIGEND_LMISC_PROTAREA, val);
	
	return result;
}

static const DWORD devids[6] = { 7, 4, 4, 4, 5, 8 };
static const DWORD  insts[6] = { 3, 3, 5, 7, 3, 3 };

static int 
KfirGeneric_ContrDownLoadMicroCode2(struct kfir_dev *kfir, int type)
{
	u32 *microbuf;
	DWORD Data;
	DWORD MicroCodeLength;
	DWORD DevId;
	int err = 0;
	DWORD i;

	printk("kfir: segment: %d  ", type);

	DevId = devids[type];
	MicroCodeLength=kfir->micro[kfir->mp];
	microbuf=kfir->micro+kfir->mp+1;
	kfir->mp+=kfir->micro[kfir->mp]+1;

 	KfirRegWrite(KFIR_DEV_INST, (DevId<<8)+1);
	KfirRegWrite(KFIR_EXEC, 0);
	KfirRegWrite(KFIR_DEV_INST, (DevId<<8)+2);
	Data = KfirRegRead(KFIR_EXEC);
	KfirRegWrite(KFIR_DEV_INST, (DevId<<8) + insts[type]);
	
	for (i = 0 ; i < MicroCodeLength ; i++) {
		KfirRegWrite(KFIR_202, (microbuf[i] >> 16) & 0xFFFF);
		KfirRegWrite(KFIR_201, microbuf[i] & 0xFFFF);
	}
	KfirRegWrite(KFIR_DEV_INST, (DevId<<8) + 1);

	KfirRegWrite(KFIR_EXEC, 0);
	KfirRegWrite(KFIR_DEV_INST, (DevId<<8) + 2);
	Data=KfirRegRead(KFIR_EXEC);

	KfirRegWrite(KFIR_DEV_INST, (DevId<<8) + insts[type] + 1);
	
	for (i = 0 ; i < MicroCodeLength ; i++) {
		Data = KfirRegRead(KFIR_201);
		Data = (KfirRegRead(KFIR_202) << 16) | (Data & 0xffff);
		if (Data != microbuf[i]) {
			err = -1;
		}
	}
	
	printk(!err ? "Done\n" : "Failed\n");
	return err;
}

static int KfirLoadMicro(struct kfir_dev *kfir)
{
	volatile DWORD Data;
	int ret=0, i;

	kfir->micro=vtv;

	if (!kfir->DVB_MODE) {
		switch (kfir->Params.StreamType) {
		default:
		case VT_KFIR_VIDEO_ES:
		case VT_KFIR_PROGRAM:
			kfir->micro=vtv;
			break;
		case VT_KFIR_TRANSPORT:
			kfir->micro=vtv2;
			break;
		case VT_KFIR_SYSTEM:
			kfir->micro=vtv1;
			break;
		}
	}
	
	if (kfir->DVB_MODE == 1) {
		switch (kfir->Params.StreamType) {
		case VT_KFIR_VIDEO_ES:
		case VT_KFIR_PROGRAM:
			kfir->micro=vtv8;
			break;
		default:
		case VT_KFIR_TRANSPORT:
			kfir->micro=vtv28;
			break;
		case VT_KFIR_SYSTEM:
			kfir->micro=vtv18;
			break;
		}
	}

	KfirRegWrite(KFIR_DEV_INST, (7<<8) + 6);
	Data = KfirRegRead(KFIR_EXEC);
	
	KfirRegWrite(KFIR_DEV_INST, (7<<8) + 5);
	Data = KfirRegRead(KFIR_EXEC);
	
	kfir->mp=0;
	for (i=0; i<6; i++) {
		if ((ret=KfirGeneric_ContrDownLoadMicroCode2(kfir, i))<0) {
			return ret;
		}
	}

	return ret;
}

int KfirGeneric_Init(struct kfir_dev *kfir)
{
	int eRet = 0;
	DWORD i;

	if (!kfir->FlFirstTime) {
		// Reset all Kfir registers
		for (i = 0 ; i < 256 ; i++) 
			KfirRegWrite(i * 4, 0);

		if (KfirLoadMicro(kfir)<0) return -1;
	}

	KfirGeneric_StartNDownLoadParams(kfir);

	/*	
	if (kfir->FlFirstTime == 0) {
	        if (Kfir_CheckDRAM(kfir) < 0) {
			return -1;
		}
	}
	*/
	kfir->FlFirstTime = 1;
	kfir->FlValid = 1;
	
	return eRet;
}


static void InitIRQs(struct kfir_dev *kfir)
{
	u32 val;

	/* Disable PLX PCI interrupt */
	val=rplxl(PLX9054_INT_CNTRL_STS);
	val &= ~PCI_INT_ENABLE;
	wplxl(PLX9054_INT_CNTRL_STS, val);

	/* Disable all Altera interrupts */
	val=raltl(ALT_ADR_INT_MASK);
	waltl(ALT_ADR_INT_MASK, val&(~ALT_D_INT_CLEAR_ALL));

	/* Clear Altera interrupts */
	val=raltl(ALT_ADR_INT_REG_CLR);
	waltl(ALT_ADR_INT_REG_CLR, val|ALT_D_INT_CLEAR_ALL);

	/* Enable PLX PCI and local (Altera) interrupt */
	val=rplxl(PLX9054_INT_CNTRL_STS);
	val |= (PCI_INT_ENABLE | LOCAL_INT_INPUT_ENABLE);
	val &= ~RETRY_ABORT_ENABLE;
	wplxl(PLX9054_INT_CNTRL_STS, val);
	
	/* Enable Altera Kfir interrupts */
	waltl(ALT_ADR_INT_MASK, ALT_D_INT_REQ_CLR_KFIR_INT);
}

static void
DecoderInit(struct kfir_dev *kfir)
{
	u32 val;

	decoder_command(kfir, DECODER_GET_STATUS, &val);
	printk("kfir: decoder status=%x\n", val);
	val=vidinmode;
	decoder_command(kfir, DECODER_SET_NORM, &val);
	val=vidinput ? 7 : 0;
	decoder_command(kfir, DECODER_SET_INPUT, &val);
}

static int 
InitBoard(struct kfir_dev *kfir)
{
	u32 val;

	InitParams(kfir);

	/* Disable PLX PCI interrupt */
	val=rplxl(PLX9054_INT_CNTRL_STS);
	val &= ~PCI_INT_ENABLE;
	wplxl(PLX9054_INT_CNTRL_STS, val);

	AlteraLoadMicro(kfir);
	Kfir_Reset(kfir);

	ADSP_KfInit(kfir);
	KfirGeneric_Init(kfir);

	init_i2c(kfir);
	scani2cbus(kfir);
	DecoderInit(kfir);

	InitIRQs(kfir);

	return 0;
}


static int kfir_open(struct video_device *dev, int flags)
{
	struct kfir_dev *kfir = (struct kfir_dev *) dev;

	kfir->recording=0;
	MOD_INC_USE_COUNT;
	return 0;   
}


static void kfir_close(struct video_device *dev)
{
	struct kfir_dev *kfir = (struct kfir_dev *) dev;

	kfir->recording=2;
	ADSP_SendCommand(kfir, DSPAPI_STOPRECORD);
	Kfir_Stop(kfir);

	{
		int i;
		int count=0;
		u16 stat;
		u32 dreq;
		while (kfir->recording) {
			if ((count++) > 5000)
				break;
			mdelay(1);
		}
		count=0;

		
		while ((stat=0xffff&KfirRegRead(KFIR_ADR_EREADY))!=1) {
			dreq=KfirRegRead(KFIR_205);
			if (!(dreq&2)) 
				for (i=0; i<64; i++)
					KfirRegRead(KFIR_203);
			if ((count++) > 1000)
				break;
			mdelay(1);
		}
		
		printk("eready=%08x\n",KfirRegRead(KFIR_ADR_EREADY));
	}
	MOD_DEC_USE_COUNT;  
}

static unsigned int kfir_poll(struct video_device *dev, struct file *file,
			      poll_table *wait)
{
        //struct kfir_dev *kfir=(struct kfir_dev *) dev;
        unsigned int mask=0;

        return mask;
}

static int kfir_mmap(struct video_device *dev, const char *adr, unsigned long size)
{
	//struct kfir_dev * kfir = (struct kfir_dev *)dev;

	return 0;
}

static struct video_channel kfir_channels[2] = {
        { 0, "Composite", 0, 0, 1, 0},
        { 1, "S-Video", 0, 0, 1, 0},
};

static int kfir_ioctl(struct video_device *dev, unsigned int cmd, void *arg)
{
	struct kfir_dev * kfir = (struct kfir_dev *) dev;
	
		switch (cmd)
	{

        case VIDIOCGCAP:
        {
                struct video_capability b;

                strcpy(b.name, &dev->name[0]);
			
                b.type 		= kfir->video.type;
						
                b.channels	= 2;	
                b.audios	= 1;
                b.maxwidth	= 720;
                b.maxheight	= 576;
                b.minwidth	= 352;
                b.minheight	= 288;
                
                if(copy_to_user(arg,&b,sizeof(b)))
                        return -EFAULT;
                return 0;
        }

        case VIDIOCGCHAN:
        {
                struct video_channel v;
                
                dprintk(KERN_ERR "dvb: VIDIOCGCHAN called\n");
                
                if(copy_from_user(&v, arg,sizeof(v)))
                        return -EFAULT;
                
                memcpy(&v, &kfir_channels[v.channel], sizeof(v));
                
                if(copy_to_user(arg,&v,sizeof(v)))
                        return -EFAULT;
                
                return 0;
                
        }

        case VIDIOCSCHAN:
        {
                struct video_channel v;
		if(copy_from_user(&v, arg,sizeof(v)))
			return -EFAULT;
		
                
		if (v.channel>0)
			return -EINVAL;
		if (v.norm > 1)
			return -EOPNOTSUPP;

                //SetMode(dvb, v.norm);

		return 0;
        }
                
        case VIDIOCGTUNER:
		return -EINVAL;

        case VIDIOCSTUNER:
		return -EINVAL;

        case VIDIOCGPICT:
        {
                struct video_picture p;
                
		if(copy_to_user(arg, &p, sizeof(struct video_picture)))
                        return -EFAULT;
                
                return 0;
        }

        case VIDIOCSPICT:
        {
                struct video_picture p;
                
                if(copy_from_user(&p, arg, sizeof(struct video_picture)))
                        return -EFAULT;
                
                //set
                
                return 0;
                
        }

        case VIDIOCSWIN:
		return -EINVAL;

        case VIDIOCGWIN:
		return -EINVAL;

        case VIDIOCCAPTURE:
		return -EINVAL;

        case VIDIOCGFBUF:
		return -EINVAL;

        case VIDIOCSFBUF:
		return -EINVAL;

        case VIDIOCKEY:
		return -EINVAL;

        case VIDIOCGFREQ:
		return -EINVAL;

        case VIDIOCSFREQ:
		return -EINVAL;

        case VIDIOCGAUDIO:
        {
                struct video_audio v;
                
                v.flags = VIDEO_AUDIO_MUTABLE;
                return 0;
        }
		/* Audio source, mute etc */
        case VIDIOCSAUDIO:
        {
          //struct video_audio v;
                
                
                return 0;
        }

	case VIDIOCSYNC:
		return -EINVAL;

        case VIDIOCMCAPTURE:
		return -EINVAL;
                
		/* Memory map buffer info */
        case VIDIOCGMBUF:
		return -EINVAL;


        case VIDIOCGUNIT:
        {
                struct video_unit vu;
                dprintk(KERN_ERR "dvb: VIDIOCGUNIT called\n");
                
                if(copy_to_user((void *)arg, (void *)&vu, sizeof(vu)))
                        return -EFAULT;
                return 0;
        }
		
		/* Get frame buffer */
        case VIDIOCGCAPTURE:
        {
                dprintk(KERN_ERR "dvb: VIDIOCGCAPTURE called\n");
                return 0;
        }
		/* Set frame buffer - root only */
        case VIDIOCSCAPTURE:
        {
                dprintk(KERN_ERR "dvb: VIDIOCSCAPTURE called\n");
                return 0;
        }

        case VIDIOCGKFIRPARAMS:
	{
                if(copy_to_user(arg, &kfir->Params, sizeof(VT_KFIR_CParams)))
                        return -EFAULT;
		return 0;
	}

        case VIDIOCSKFIRPARAMS:
	{
                if(copy_from_user(&kfir->Params, arg, sizeof(VT_KFIR_CParams)))
                        return -EFAULT;
		return 0;
	}

        default:
                return -ENOIOCTLCMD;
	}
	return 0;
}        


static long kfir_write(struct video_device *dev, const char *buf,
		       unsigned long count, int nonblock)
{
	/* not supported by Kfir1 */
        return 0;
}


static long record(struct kfir_dev *kfir, char *buf,
		   unsigned long count, int nonblock)
{
	DECLARE_WAITQUEUE(wait, current);
        unsigned long todo = count;
        int blocksize, split;
        
        if (!kfir->recording)
        {
		dprintk("kfir: start recording\n");
		///InitIRQs(kfir);
		KfirGeneric_StartNDownLoadParams(kfir);
		kfir->readp=kfir->writep=0;
		kfir->recording=1;
		kfir->IntVideoCnt=0;
		waltl(ALT_ADR_CONTROL_1, 2);
		ADSP_Start(kfir);
		KfirRegRead(205*4);
		Kfir_Start(kfir);
        }
        
        while (todo > 0) {
                if (kfir->readp == kfir->writep) {
                        if (nonblock)
                                return count-todo;
                        add_wait_queue(&kfir->wq, &wait);
                        while (1) {
                                set_current_state(TASK_INTERRUPTIBLE);
                                if (kfir->readp != kfir->writep
                                    || signal_pending(current))
                                        break;
                                schedule();
                        }
                        current->state=TASK_RUNNING;
                        remove_wait_queue(&kfir->wq, &wait);
                        if (kfir->readp==kfir->writep) 
                                return 0;
                }

                split=MPGBUFSIZE;
                blocksize=kfir->writep-kfir->readp;
                if (blocksize < 0) {
                        blocksize+=MPGBUFSIZE;
                        split=MPGBUFSIZE-kfir->readp;
                }
                if (blocksize > todo)
                        blocksize=todo;
                if (split < blocksize) {
                        if (copy_to_user(buf, kfir->mpgbuf+kfir->readp, split))
				return -EFAULT;
                        buf+=split;
                        kfir->readp=0;
                        todo-=split;
                        blocksize-=split;
                }
                if (blocksize) {
                        if (copy_to_user(buf, kfir->mpgbuf+kfir->readp, blocksize))
                                return -EFAULT;
                        kfir->readp = (kfir->readp + blocksize) 
				& (MPGBUFSIZE-1);
                        todo-=blocksize;
                        buf+=blocksize;
                }
        }
	return count;
}


static long kfir_read(struct video_device *dev, char *buf,
		      unsigned long count, int nonblock)
{
	struct kfir_dev * kfir = (struct kfir_dev *) dev;

	return record(kfir, buf, count, nonblock);
}

static int kfir_init_done(struct video_device *dev)
{
	return 0;
}


/* template for video_device-structure */
static struct video_device kfir_template = {
	"Kfir",
        VID_TYPE_MPEG_ENCODER,  
        VID_HARDWARE_SAA7146,  //FIXME: need to get VID_HARDWARE_KFIR registered
        kfir_open,
        kfir_close,
        kfir_read,
	kfir_write,		
	kfir_poll,
	kfir_ioctl,
	kfir_mmap,
	kfir_init_done,
	NULL,
	0,
	0
};

static int kfir_init_device (struct pci_dev *pdev)
{
	u32 class_rev;
        struct kfir_dev *kfir;

	if (num_kfir>=MAX_NUM_KFIR)
		return -1;

	kfir=&kfirs[num_kfir];
	num_kfir++;
#if LINUX_VERSION_CODE >= 0x020300
	pdev->driver_data=kfir;
#endif
	memset(kfir, 0, sizeof(struct kfir_dev));

	kfir->pdev=pdev;
	pci_read_config_dword(kfir->pdev, PCI_CLASS_REVISION, &class_rev);
	kfir->revision=class_rev&0xff;
	kfir->kfirreset=0;
	
	pci_read_config_word(kfir->pdev, PCI_SUBSYSTEM_ID, &kfir->subsystem_device);

	if (kfir->subsystem_device == 0x200)
		kfir->mode = KFIR_GENERIC_MODE;
	else
		kfir->mode = KFIR_MASTER_MODE;
	kfir->ALT_ADR_INT_MASK = ALT_ADR_INT_MASK_REV_5E;
	kfir->ALT_ADR_INT_REG_CLR = ALT_ADR_INT_REG_CLR_REV_5E;
	kfir->ALT_ADR_KFIR_AUDIO_BASE = 0xFFFFFFFF;
	kfir->ALT_ADR_RESET_REG = ALT_ADR_RESET_REG_REV_5E;

	kfir->ALT_ADR_ADSP_ADDR = ALT_ADR_ADSP_ADDR_REV_5E;
	kfir->ALT_ADR_ADSP_DATA = ALT_ADR_ADSP_DATA_REV_5E;
	kfir->ALT_ADR_ADSP_IRQ = ALT_ADR_ADSP_IRQ_REV_5E;
	kfir->ALT_ADR_ADSP_FREQ = ALT_ADR_AUDIO_FREQ_REV_5E;
	kfir->ALT_ADR_I2C = ALT_ADR_I2C_REV_5E;
	kfir->ALT_ADR_CONTROL_0 = 0xFFFFFFFF;
	kfir->ALT_ADR_CONTROL_1 = ALT_ADR_OUTPUT_REV_5E;
	kfir->ALT_ADR_CLOCK_PADI = ALT_ADR_CLOCK_PADI_DEF;

#if LINUX_VERSION_CODE >= 0x020300
	kfir->mem0=ioremap(pdev->resource[0].start&PCI_BASE_ADDRESS_MEM_MASK, 0x100);
	kfir->mem1=(char *)(pdev->resource[1].start&PCI_BASE_ADDRESS_MEM_MASK);
	kfir->mem2=ioremap(pdev->resource[2].start&PCI_BASE_ADDRESS_MEM_MASK, 0x200000);
	kfir->mem3=ioremap(pdev->resource[3].start&PCI_BASE_ADDRESS_MEM_MASK, 0x400);
#else
	kfir->mem0=ioremap(pdev->base_address[0]&PCI_BASE_ADDRESS_MEM_MASK, 0x100);
	kfir->mem1=(char *)(pdev->base_address[1]&PCI_BASE_ADDRESS_MEM_MASK);
	kfir->mem2=ioremap(pdev->base_address[2]&PCI_BASE_ADDRESS_MEM_MASK, 0x200000);
	kfir->mem3=ioremap(pdev->base_address[3]&PCI_BASE_ADDRESS_MEM_MASK, 0x400);
#endif

	kfir->AlteraReg=kfir->mem2;
	kfir->KfirReg=kfir->AlteraReg;
	kfir->KfirReg+=0x2000;

	printk("kfir: Kfir board revision %02x\n", (int) kfir->revision);

	kfir->dmabuf=kmalloc(MUX_DMA_LENGTH, GFP_KERNEL);
	if (!kfir->dmabuf) 
		return -1;

	kfir->mpgbuf=vmalloc(MPGBUFSIZE);
	if (!kfir->mpgbuf) {
		kfree(kfir->dmabuf);
		return -1;
	}


        /* Copy structure */
        memcpy(&kfir->video, &kfir_template, sizeof(struct video_device));
        
        if(0 > (video_register_device(&kfir->video, VFL_TYPE_GRABBER))) {
                printk(KERN_ERR "kfir: can't register videodevice\n");
                return -1;
        }
        init_waitqueue_head(&kfir->wq);

	init_plx9054(kfir);

	return InitBoard(kfir);
}


static void kfir_remove_device(struct kfir_dev *kfir)
{
	DWORD val;

        printk("kfir: kfir_remove_one()\n");

	val=raltl(ALT_ADR_INT_MASK);
	waltl(ALT_ADR_INT_MASK, val&(~ALT_D_INT_CLEAR_ALL));

	val=raltl(ALT_ADR_INT_REG_CLR);
	waltl(ALT_ADR_INT_REG_CLR, val|ALT_D_INT_CLEAR_ALL);

	if (kfir->mode == KFIR_GENERIC_MODE) { 
		val=rplxl(DMA0_DONE_ENABLE);
		val&=~DMA0_DONE_ENABLE;
		wplxl(DMA0_DONE_ENABLE, val);
		
		val=rplxl(PLX9054_INT_CNTRL_STS);
		val |= DMA0_ENABLE;
		wplxl(PLX9054_INT_CNTRL_STS, val);
		val |= DMA0_START;
		wplxl(PLX9054_INT_CNTRL_STS, val);
		val &= ~DMA0_ENABLE;
		wplxl(PLX9054_INT_CNTRL_STS, val);
		val |= DMA0_ABORT;
		wplxl(PLX9054_INT_CNTRL_STS, val);
		val &= ~DMA0_CLEAR_INT;
		wplxl(PLX9054_INT_CNTRL_STS, val);
	}

        if(kfir->video.minor!=-1)
                video_unregister_device(&kfir->video);

	i2c_bit_del_bus(&kfir->i2c_adap);

	iounmap(kfir->mem0);
	iounmap(kfir->mem2);
	iounmap(kfir->mem3);

	free_irq(kfir->pdev->irq, kfir);

	if (kfir->dmabuf)
		kfree(kfir->dmabuf);
	if (kfir->mpgbuf)
		vfree(kfir->mpgbuf);
}


#if LINUX_VERSION_CODE < 0x020300

static int find_kfir(void) 
{
	struct pci_dev *dev = NULL;
	int result = 0;
	
	if (!pci_present()) {
		printk(KERN_ERR "kfir: PCI-BIOS not present or not accessable!\n");
		return -1;
	}
        while ((dev = pci_find_device(PCI_VENDOR_ID_VISIONTECH,
				      PCI_DEVICE_ID_VISIONTECH_KFIR_5E, dev))) 
	{
		if (num_kfir) 
			ddelay(100);
		result = kfir_init_device(dev);

                if (result<0)
                        return result;
	}
	if (num_kfir)
		printk(KERN_INFO "kfir: %d kfir card(s) found.\n", num_kfir);

	return num_kfir;
}

#ifdef MODULE

int init_module(void) 
{
	if (0 > find_kfir())
		return -EIO;
	return 0;
}

void cleanup_module(void) 
{
	int i;
	
	for (i=0; i<num_kfir; i++)
		kfir_remove_device(&kfirs[i]);
}
#endif


#else

static int __devinit kfir_init_one (struct pci_dev *pdev,
		const struct pci_device_id *ent)
{
	return kfir_init_device(pdev);
}

static void __devexit kfir_remove_one (struct pci_dev *pdev)
{
	return kfir_remove_device((struct kfir_dev *) pdev->driver_data);
}

static void kfir_suspend(struct pci_dev *pdev)
{
        printk("kfir: kfir_suspend()\n");
}

static void kfir_resume(struct pci_dev *pdev)
{
        printk("kfir: kfir_resume()\n");
}


static struct pci_device_id kfir_pci_tbl[] __devinitdata = {
	{ PCI_VENDOR_ID_VISIONTECH, PCI_DEVICE_ID_VISIONTECH_KFIR_5E,
	  PCI_ANY_ID, PCI_ANY_ID, },
	{ 0,},
};

MODULE_DEVICE_TABLE(pci, kfir_pci_tbl);
	
static struct pci_driver kfir_driver = {
	name:		"kfir",
	id_table:	kfir_pci_tbl,
	probe:		kfir_init_one,
	remove:		kfir_remove_one,
	suspend:	kfir_suspend,
	resume:		kfir_resume,
};

int __init kfir_init_module(void)
{
	return pci_module_init(&kfir_driver);
}

void __exit kfir_cleanup_module(void)
{
	pci_unregister_driver(&kfir_driver);
}

module_init(kfir_init_module);
module_exit(kfir_cleanup_module);

#endif 

/*
 * Local variables:
 *  c-indent-level: 8
 *  c-basic-offset: 8
 *  tab-width: 8
 * End:
 */
